# pylint: disable=no-member, not-callable, missing-docstring, line-too-long, invalid-name
from os import mkdir
from os.path import isdir, isfile, join
from shutil import rmtree

import numpy as np
import pymatgen
import torch
from torch.utils.data import Dataset
from tqdm import tqdm


class CrystalCIF(Dataset):
    """
    root
    |_______index.npy
    |_______names.npy
    |_______property0.pth
    |_______property1.pth
    |_______cif
    |        |____name0.cif
    |        |____name1.cif
    |
    |_______preprocessed (script creates this folder and its contents)
             |____geometries.pth
             |____atomic_charges.pth
             |____lattice_params.pth
             |____max_radius_{max_radius_value}
                    |_____radii.pth
                    |_____bs.pth
                    |_____partitions.pth
    """
    def __init__(self, root, max_radius, material_properties=None, bs_pad=64):
        """
        :param root: string, path to the root directory of dataset
        :param max_radius: float, radius of sphere
        :param material_properties: (optional) list of file paths containing additional properties, one type of property per file
        :param bs_pad: integer, length of fixed size vector in which we embed bs (single list)
        """
        preprocessed_dir = join(root, 'preprocessed')
        preprocessed_radius_dir = join(preprocessed_dir, f'max_radius_{max_radius}')

        if isdir(preprocessed_radius_dir) and not all(isfile(join(preprocessed_radius_dir, f)) for f in ['radii.pth', 'bs.pth', 'partitions.pth']):
            rmtree(preprocessed_radius_dir)
            CrystalCIF.preprocess(root, max_radius, bs_pad)
        elif not isdir(preprocessed_radius_dir):
            CrystalCIF.preprocess(root, max_radius, bs_pad)
        else:
            pass

        self.names = np.load(join(root, 'names.npy'))
        self.size = len(self.names)
        self.geometries = torch.load(join(preprocessed_dir, 'geometries.pth'))
        self.atomic_charges = torch.load(join(preprocessed_dir, 'atomic_charges.pth'))
        self.lattice_params = torch.load(join(preprocessed_dir, 'lattice_params.pth'))

        self.radii = torch.load(join(preprocessed_radius_dir, 'radii.pth'))
        self.bs = torch.load(join(preprocessed_radius_dir, 'bs.pth'))
        self.partitions = torch.load(join(preprocessed_radius_dir, 'partitions.pth'))

        if material_properties:
            self.properties = torch.stack([torch.load(join(root, property_path)) for property_path in material_properties], dim=1)
        else:
            self.properties = None

    def __getitem__(self, item_id):
        properties = None if self.properties is None else self.properties[item_id]
        radii_start, radii_end, bs_start, bs_end = self.partitions[item_id]
        return self.names[item_id], self.radii[radii_start:radii_end], self.bs[bs_start:bs_end], self.geometries[item_id], \
            self.atomic_charges[item_id], self.lattice_params[item_id], properties

    def __len__(self):
        return self.size

    @staticmethod
    def preprocess(root, max_radius=None, bs_pad=64):
        """
        Allows calls without class instance: CrystalCIF.preprocess(...).
        :param root: string, path to the root directory of dataset
        :param max_radius: float, (optional) radius of sphere
        :param bs_pad: integer, (unused for max_radius=None) length of fixed size vector in which we embed bs (single list)

        Notes:
            Let there be bs_lists [17, 18, 19, 24, 25] and [1, 2, 5, 6], then with bs_pad = 8 they will be stacked in tensor as follows:
            [[5, 17, 18, 19, 24, 25, 0, 0],
             [4,  1,  2,  5,  6,  0, 0, 0]]
            where first element in each row is a number n of relevant elements in it, next n elements are those copied from the lists,
            and rest padded with zeros for length to match bs_pad.

            Pros (over list of lists): has smaller overhead from structure meta-data, occupies contingent memory area.
            Cons (over list of lists): additionally stores lengths and pad elements.
        """
        # region 0. Set up
        preprocessed_dir = join(root, 'preprocessed')
        if not isdir(preprocessed_dir):
            mkdir(preprocessed_dir)

        if max_radius:
            max_radius_dir = join(preprocessed_dir, f'max_radius_{max_radius}')
            if not isdir(max_radius_dir):
                mkdir(max_radius_dir)
        # endregion

        # region 1. Init
        index = np.load(join(root, 'index.npy'))

        # TODO: preallocate memory for tensors and write into it, instead of using lists.
        #  Can save ~30% on peak memory, but requires proper reallocation on the go for radii as shape isn't known beforehand.

        site_a_coords_list = []
        atomic_charges_list = []
        lattice_params_list = []

        if max_radius:
            bs_list = []
            radii_list = []
            partitions_list = []
        # endregion

        # region 2. Process
        radii_partition_start = 0
        bs_partition_start = 0
        for file_rel_path in tqdm(index):
            structure = pymatgen.Structure.from_file((join(root, 'cif', file_rel_path)))

            site_a_coords_entry = torch.stack([torch.from_numpy(site.coords) for site in structure.sites])
            atomic_charges_entry = torch.tensor([atom.number for atom in structure.species])

            site_a_coords_list.append(site_a_coords_entry)
            atomic_charges_list.append(atomic_charges_entry)
            lattice_params_list.append(structure.lattice.abc + structure.lattice.angles)

            if max_radius:
                radii_proxy_list = []
                bs_proxy_list = []
                for site_a_coords in site_a_coords_entry:
                    nei = structure.get_sites_in_sphere(site_a_coords.numpy(), max_radius, include_index=True)
                    assert nei, f"Encountered empty nei for {file_rel_path}: {site_a_coords}"

                    # save storage/RAM, transfer of small parts to the long "register" (for indexing) on the fly is comp. cheap
                    bs_entry = np.zeros(bs_pad, dtype=np.short)
                    bs_entry_data = [entry[2] for entry in nei]
                    bs_entry_data_len = len(bs_entry_data)
                    assert bs_entry_data_len < bs_pad, f"Encountered bs vector ({bs_entry_data_len}) longer than provided bs_pad ({bs_pad})"

                    bs_entry[0] = bs_entry_data_len                                                 # store number of meaningful entries as a first element
                    bs_entry[1:1 + bs_entry_data_len] = bs_entry_data                                 # store said meaningful entries in consecutive cells
                    bs_proxy_list.append(bs_entry)

                    site_b_coords = np.array([entry[0].coords for entry in nei])                    # [r_part_a, 3]
                    site_a_coords = np.array(site_a_coords).reshape(1, 3)                           # [1, 3]
                    radii_proxy_list.append(site_b_coords - site_a_coords)                          # implicit broadcasting of site_a_coords

                radii_proxy = np.concatenate(radii_proxy_list)                                      # [r, 3]
                radii_proxy[np.linalg.norm(radii_proxy, ord=2, axis=-1) < 1e-10] = 0.
                radii_list.append(radii_proxy)

                bs_list.append(bs_proxy_list)

                partitions_list.append((radii_partition_start, radii_partition_start + radii_proxy.shape[0],
                                        bs_partition_start, bs_partition_start + len(bs_proxy_list)))
                radii_partition_start += radii_proxy.shape[0]
                bs_partition_start += len(bs_proxy_list)
        # endregion

        # region 3. Post-process
        lattice_params = torch.tensor(lattice_params_list)
        del lattice_params_list

        if max_radius:
            bs = torch.from_numpy(np.concatenate(bs_list))
            del bs_list

            radii = torch.from_numpy(np.concatenate(radii_list))
            del radii_list

            partitions = torch.tensor(partitions_list, dtype=torch.long)
            del partitions_list
        # endregion

        # region 4. Store
        torch.save(site_a_coords_list, join(preprocessed_dir, 'geometries.pth'))                    # list of z tensors [a_i, 3]            - xyz
        torch.save(atomic_charges_list, join(preprocessed_dir, 'atomic_charges.pth'))               # list of z tensors [a_i]
        torch.save(lattice_params, join(preprocessed_dir, 'lattice_params.pth'))                    # [z, 6]                                - xyz and angles (in degrees)
        del site_a_coords_list, atomic_charges_list, lattice_params

        if max_radius:
            torch.save(bs, join(max_radius_dir, 'bs.pth'))                                          # tensor [sum(a_i), bs_pad]
            del bs

            torch.save(radii, join(max_radius_dir, 'radii.pth'))                                    # tensor [sum(r_i), 3]                  - xyz
            del radii

            # tensor [z, 4]                         - start/end of radii slice and bs slice corresponding to z
            torch.save(partitions, join(max_radius_dir, 'partitions.pth'))
            del partitions
        # endregion



# pylint: disable=arguments-differ, redefined-builtin, missing-docstring, line-too-long, no-member, invalid-name, abstract-method
import math
from functools import partial

import torch

from e3nn import rsh, o3
from e3nn.kernel_mod import FrozenKernel
from e3nn.non_linearities.rescaled_act import swish
from e3nn.radial import CosineBasisModel


class Convolution(torch.nn.Module):
    def __init__(self, Rs_in, Rs_out, size, steps=(1, 1, 1), lmax=None, fuzzy_pixels=False, allow_unused_inputs=False, allow_zero_outputs=False, **kwargs):
        super().__init__()

        r = torch.linspace(-1, 1, size)
        x = r * steps[0] / min(steps)
        x = x[x.abs() <= 1]
        y = r * steps[1] / min(steps)
        y = y[y.abs() <= 1]
        z = r * steps[2] / min(steps)
        z = z[z.abs() <= 1]
        r = torch.stack(torch.meshgrid(x, y, z), dim=-1)  # [x, y, z, R^3]

        R = partial(CosineBasisModel, max_radius=1.0, number_of_basis=(size + 1) // 2, h=50, L=3, act=swish)
        self.kernel = FrozenKernel(
            Rs_in,
            Rs_out,
            R,
            r,
            selection_rule=partial(o3.selection_rule_in_out_sh, lmax=lmax),
            normalization='component',
            allow_unused_inputs=allow_unused_inputs,
            allow_zero_outputs=allow_zero_outputs,
        )
        self.kwargs = kwargs

        if fuzzy_pixels:
            # re-evaluate spherical harmonics by adding randomness
            r = r.reshape(-1, 3)
            r = r[self.kernel.radii > 0]
            rand = torch.rand(20**3, *r.shape).mul(2).sub(1)  # [-1, 1]
            rand.mul_(1 / (size - 1))
            rand[:, :, 0].mul_(steps[0] / min(steps))
            rand[:, :, 1].mul_(steps[1] / min(steps))
            rand[:, :, 2].mul_(steps[2] / min(steps))
            r = rand + r.unsqueeze(0)  # [rand, batch, R^3]
            Y = rsh.spherical_harmonics_xyz([(1, l, p) for _, l, p in self.kernel.Rs_f], r)
            # Y  # [rand, batch, l_filter * m_filter]
            Y.mul_(math.sqrt(4 * math.pi))  # normalization='component'
            self.kernel.Y.copy_(Y.mean(0))

    def forward(self, features):
        """
        :param tensor features: [batch, x, y, z, channel]
        :return: [batch, x, y, z, channel]
        """
        features = torch.einsum('txyzi->tixyz', features)
        k = torch.einsum('xyzij->ijxyz', self.kernel())
        k.mul_(1 / k[0, 0].numel() ** 0.5)
        features = torch.nn.functional.conv3d(features, k, **self.kwargs)
        features = torch.einsum('tixyz->txyzi', features)
        return features



# pylint: disable=invalid-name, arguments-differ, missing-docstring, line-too-long, no-member, redefined-builtin
from e3nn import rs


class IrrepTensor():
    def __init__(self, tensor, Rs):
        Rs = rs.convention(Rs)
        if tensor.shape[-1] != rs.dim(Rs):
            raise ValueError("Last tensor dimension and Rs do not have same dimension.")
        self.tensor = tensor
        self.Rs = Rs



# pylint: disable=arguments-differ, redefined-builtin, missing-docstring, line-too-long, no-member, invalid-name
import torch
import torch.nn.functional as F


class LowPassFilter(torch.nn.Module):
    def __init__(self, scale, stride=1):
        """
        :param float scale:
        :param int stride:
        """
        super().__init__()

        sigma = 0.5 * (scale ** 2 - 1) ** 0.5

        size = int(1 + 2 * 2.5 * sigma)
        if size % 2 == 0:
            size += 1

        rng = torch.arange(size, dtype=torch.get_default_dtype()) - size // 2  # [-(size // 2), ..., size // 2]
        x = rng.reshape(size, 1, 1).expand(size, size, size)
        y = rng.reshape(1, size, 1).expand(size, size, size)
        z = rng.reshape(1, 1, size).expand(size, size, size)

        kernel = torch.exp(- (x ** 2 + y ** 2 + z ** 2) / (2 * sigma ** 2))
        kernel = kernel / kernel.sum()
        kernel = kernel.reshape(1, 1, size, size, size)
        self.register_buffer('kernel', kernel)

        self.scale = scale
        self.stride = stride
        self.size = size

    def forward(self, image):
        """
        :param tensor image: [..., x, y, z, channel]
        """
        if self.scale <= 1:
            assert self.stride == 1
            return image

        out = image.reshape(-1, *image.shape[-4:])
        out = torch.einsum('txyzi->tixyz', out)
        out = out.reshape(-1, 1, *out.shape[-3:])
        out = F.conv3d(out, self.kernel, padding=self.size // 2, stride=self.stride)
        out = out.reshape(-1, image.shape[-1], *out.shape[-3:])
        out = torch.einsum('tixyz->txyzi', out)
        out = out.reshape(*image.shape[:-4], *out.shape[-4:])
        return out



# pylint: disable=not-callable, no-member, invalid-name, line-too-long, unexpected-keyword-arg, too-many-lines, import-outside-toplevel
"""
Some functions related to SO3 and his usual representations

Using ZYZ Euler angles parametrisation
"""
import gc
import math
import os
from functools import lru_cache
from typing import Callable, List, Tuple

import lie_learn.representations.SO3.pinchon_hoggan.pinchon_hoggan_dense as ph
import scipy
import scipy.linalg
import torch
from lie_learn.representations.SO3.wigner_d import wigner_D_matrix

from e3nn.util.cache_file import cached_picklesjar
from e3nn.util.default_dtype import torch_default_dtype


def rot_z(gamma, dtype=None, device=None):
    """
    Rotation around Z axis
    """
    if dtype is None:
        dtype = torch.get_default_dtype()
    if not torch.is_tensor(gamma):
        gamma = torch.tensor(gamma, dtype=dtype, device=device)
    else:
        gamma = gamma.to(dtype=dtype, device=device)

    return torch.stack([
        torch.stack([gamma.cos(),
                     -gamma.sin(),
                     gamma.new_zeros(gamma.shape)], dim=-1),
        torch.stack([gamma.sin(),
                     gamma.cos(),
                     gamma.new_zeros(gamma.shape)], dim=-1),
        torch.stack([gamma.new_zeros(gamma.shape),
                     gamma.new_zeros(gamma.shape),
                     gamma.new_ones(gamma.shape)], dim=-1)
    ], dim=-2)


def rot_y(beta, dtype=None, device=None):
    """
    Rotation around Y axis
    """
    if dtype is None:
        dtype = torch.get_default_dtype()
    if not torch.is_tensor(beta):
        beta = torch.tensor(beta, dtype=dtype, device=device)
    else:
        beta = beta.to(dtype=dtype, device=device)

    return torch.stack([
        torch.stack([beta.cos(),
                     beta.new_zeros(beta.shape),
                     beta.sin()], dim=-1),
        torch.stack([beta.new_zeros(beta.shape),
                     beta.new_ones(beta.shape),
                     beta.new_zeros(beta.shape)], dim=-1),
        torch.stack([-beta.sin(),
                     beta.new_zeros(beta.shape),
                     beta.cos()], dim=-1),
    ], dim=-2)


# The following two functions (rot and xyz_to_angles) satisfies that
# rot(*xyz_to_angles([x, y, z]), 0) @ np.array([[0], [0], [1]])
# is proportional to
# [x, y, z]

def rot(alpha, beta, gamma, dtype=None, device=None):
    """
    ZYZ Euler angles rotation
    """
    return rot_z(alpha, dtype, device) @ rot_y(beta, dtype, device) @ rot_z(gamma, dtype, device)


def rand_rot():
    """
    random rotation matrix
    """
    return rot(*rand_angles())


def rand_angles():
    """
    random rotation angles
    """
    alpha, gamma = 2 * math.pi * torch.rand(2)
    beta = torch.rand(()).mul(2).sub(1).acos()
    return alpha, beta, gamma


def angles_to_xyz(alpha, beta):
    """
    Convert (alpha, beta) into point (x, y, z) on the sphere
    """
    x = torch.sin(beta) * torch.cos(alpha)
    y = torch.sin(beta) * torch.sin(alpha)
    z = torch.cos(beta)
    return torch.stack([x, y, z], dim=-1)


def xyz_to_angles(x, y=None, z=None):
    """
    Convert point (x, y, z) on the sphere into (alpha, beta)
    """
    if y is not None and z is not None:
        x = torch.stack([x, y, z], dim=-1)

    x = torch.nn.functional.normalize(x, p=2, dim=-1)  # forward 0's instead of nan for zero-radius
    x.masked_fill_(x < -1., -1.)                       # mitigate numerical inaccuracies from normalization
    x.masked_fill_(x > 1., 1.)

    beta = torch.acos(x[..., 2])
    alpha = torch.atan2(x[..., 1], x[..., 0])
    return alpha, beta


def rot_to_abc(R):
    """
    Convert rotation matrix into (alpha, beta, gamma)
    """
    x = R @ R.new_tensor([0, 0, 1])
    a, b = xyz_to_angles(x)
    R = rot(a, b, a.new_zeros(a.shape)).transpose(-1, -2) @ R
    c = torch.atan2(R[..., 1, 0], R[..., 0, 0])
    return a, b, c


def compose(a1, b1, c1, a2, b2, c2):
    """
    (a, b, c) = (a1, b1, c1) composed with (a2, b2, c2)
    """
    comp = rot(a1, b1, c1) @ rot(a2, b2, c2)
    xyz = comp @ torch.tensor([0, 0, 1.])
    a, b = xyz_to_angles(xyz)
    rotz = rot(0, -b, -a) @ comp
    c = torch.atan2(rotz[1, 0], rotz[0, 0])
    return a, b, c


def compose_with_parity(a1, b1, c1, p1, a2, b2, c2, p2):
    """
    (a, b, c, p) = (a1, b1, c1, p1) composed with (a2, b2, c2, p2)
    """
    return compose(a1, b1, c1, a2, b2, c2) + ((p1 + p2) % 2,)


def irr_repr(order, alpha, beta, gamma, dtype=None, device=None):
    """
    irreducible representation of SO3
    - compatible with compose and spherical_harmonics
    """
    abc = [alpha, beta, gamma]
    for i, x in enumerate(abc):
        if torch.is_tensor(x):
            abc[i] = x.item()
            if dtype is None:
                dtype = x.dtype
            if device is None:
                device = x.device
    if dtype is None:
        dtype = torch.get_default_dtype()
    return torch.tensor(wigner_D_matrix(order, *abc), dtype=dtype, device=device)


def derivative_irr_repr(order, alpha, beta, gamma, dtype=None, device=None):
    """
    derivative of irreducible representation of SO3
    returns (dDda, dDdb, dDdc)
    """
    abc = [alpha, beta, gamma]
    for i, x in enumerate(abc):
        if torch.is_tensor(x):
            abc[i] = x.item()
            if dtype is None:
                dtype = x.dtype
            if device is None:
                device = x.device
    if dtype is None:
        dtype = torch.get_default_dtype()
    dDdabc = ph.derivative_rot_mat(*abc, l=order, J=ph.Jd[order])
    dDda, dDdb, dDdc = [torch.tensor(i, dtype=dtype, device=device) for i in dDdabc]
    return dDda, dDdb, dDdc


TY_SELECTION_RULE = Callable[[int, int, int, int], List[int]]


def selection_rule(l1: int, _p1: int, l2: int, _p2: int, lmax=None, lfilter=None) -> List[int]:
    """
    selection rule
    :return: list from |l1-l2|... to l1+l2
    """
    if lmax is None:
        l_max = l1 + l2
    else:
        l_max = min(lmax, l1 + l2)
    ls = list(range(abs(l1 - l2), l_max + 1))
    if lfilter is not None:
        ls = list(filter(lfilter, ls))
    return ls


def selection_rule_in_out_sh(l_in: int, p_in: int, l_out: int, p_out: int, lmax=None) -> List[int]:
    """
    all possible spherical harmonics such that
    Input * SH = Output
    """
    return [l for l in selection_rule(l_in, p_in, l_out, p_out, lmax) if p_out in [0, p_in * (-1) ** l]]


################################################################################
# Linear algebra
################################################################################

def kron(*matrices):
    """
    Kroneker product between matrices
    """
    for m in matrices:
        assert m.dim() == 2

    if len(matrices) == 0:
        return torch.ones(1, 1)
    if len(matrices) == 1:
        return matrices[0]

    x, y, *matrices = matrices
    z = torch.einsum("ij,kl->ikjl", x, y).reshape(x.size(0) * y.size(0), x.size(1) * y.size(1))

    if matrices:
        return kron(z, *matrices)
    return z


def direct_sum(*matrices):
    """
    Direct sum of matrices, put them in the diagonal
    """
    front_indices = matrices[0].shape[:-2]
    m = sum(x.size(-2) for x in matrices)
    n = sum(x.size(-1) for x in matrices)
    total_shape = list(front_indices) + [m, n]
    out = matrices[0].new_zeros(*total_shape)
    i, j = 0, 0
    for x in matrices:
        m, n = x.shape[-2:]
        out[..., i: i + m, j: j + n] = x
        i += m
        j += n
    return out


################################################################################
# 3j symbol
################################################################################

def wigner_3j(l1, l2, l3, cached=False, dtype=None, device=None, like=None):
    """
    Computes the 3-j symbol
    https://en.wikipedia.org/wiki/3-j_symbol

    D(l1)_il D(l2)_jm D(l3)_kn Q_lmn == Q_ijk
    """
    assert isinstance(l1, int)
    assert isinstance(l2, int)
    assert isinstance(l3, int)

    if dtype is None:
        if like is not None:
            dtype = like.dtype
        else:
            dtype = torch.get_default_dtype()
    if device is None:
        if like is not None:
            device = like.device
        else:
            device = 'cpu'

    # return a clone to avoid that the user modifies the matrices in-place
    if cached:
        return _cached_wigner_3j(l1, l2, l3, dtype, device).clone()
    return _wigner_3j(l1, l2, l3).to(dtype=dtype, device=device).clone()


@lru_cache(maxsize=None)
def _cached_wigner_3j(l1, l2, l3, dtype, device):
    return _wigner_3j(l1, l2, l3).to(dtype=dtype, device=device)


def _wigner_3j(l1, l2, l3):
    if l1 <= l2 <= l3:
        return __wigner_3j(l1, l2, l3)
    if l1 <= l3 <= l2:
        return __wigner_3j(l1, l3, l2).transpose(1, 2) * (-1) ** (l1 + l2 + l3)
    if l2 <= l1 <= l3:
        return __wigner_3j(l2, l1, l3).transpose(0, 1) * (-1) ** (l1 + l2 + l3)
    if l3 <= l2 <= l1:
        return __wigner_3j(l3, l2, l1).transpose(0, 2) * (-1) ** (l1 + l2 + l3)
    if l2 <= l3 <= l1:
        return __wigner_3j(l2, l3, l1).transpose(0, 2).transpose(1, 2)
    if l3 <= l1 <= l2:
        return __wigner_3j(l3, l1, l2).transpose(0, 2).transpose(0, 1)


@cached_picklesjar(os.path.join(os.path.dirname(__file__), 'cache/wigner_3j'))
def __wigner_3j(l1, l2, l3, _version=1):  # pragma: no cover
    """
    Computes the 3-j symbol
    https://en.wikipedia.org/wiki/3-j_symbol

    Closely related to the Clebsch–Gordan coefficients

    D(l1)_il D(l2)_jm D(l3)_kn Q_lmn == Q_ijk
    """
    # these three propositions are equivalent
    assert abs(l2 - l3) <= l1 <= l2 + l3
    assert abs(l3 - l1) <= l2 <= l3 + l1
    assert abs(l1 - l2) <= l3 <= l1 + l2

    def _DxDxD(a, b, c):
        D1 = irr_repr(l1, a, b, c)
        D2 = irr_repr(l2, a, b, c)
        D3 = irr_repr(l3, a, b, c)
        return torch.einsum('il,jm,kn->ijklmn', (D1, D2, D3)).reshape(n, n)

    n = (2 * l1 + 1) * (2 * l2 + 1) * (2 * l3 + 1)
    random_angles = [
        [4.41301023, 5.56684102, 4.59384642],
        [4.93325116, 6.12697327, 4.14574096],
        [0.53878964, 4.09050444, 5.36539036],
        [2.16017393, 3.48835314, 5.55174441],
        [2.52385107, 0.29089583, 3.90040975],
    ]

    with torch_default_dtype(torch.float64):
        B = torch.zeros((n, n))
        for abc in random_angles:
            D = _DxDxD(*abc) - torch.eye(n)
            B += D.T @ D
            del D
            gc.collect()

    # ask for one (smallest) eigenvalue/eigenvector pair if there is only one exists, otherwise ask for two
    s, v = scipy.linalg.eigh(B.numpy(), eigvals=(0, min(1, n - 1)), overwrite_a=True)
    del B
    gc.collect()

    kernel = v.T[s < 1e-10]
    null_space = torch.from_numpy(kernel)

    assert null_space.size(0) == 1, null_space.size()  # unique subspace solution
    Q = null_space[0]
    Q = Q.reshape(2 * l1 + 1, 2 * l2 + 1, 2 * l3 + 1)

    if next(x for x in Q.flatten() if x.abs() > 1e-10 * Q.abs().max()) < 0:
        Q.neg_()

    Q[Q.abs() < 1e-14] = 0

    with torch_default_dtype(torch.float64):
        abc = rand_angles()
        _Q = torch.einsum("il,jm,kn,lmn", (irr_repr(l1, *abc), irr_repr(l2, *abc), irr_repr(l3, *abc), Q))
        assert torch.allclose(Q, _Q)

    assert Q.dtype == torch.float64
    return Q  # [m1, m2, m3]


################################################################################
# Change of basis
################################################################################

def xyz_to_irreducible_basis(check=True):
    """
    to convert a vector [x, y, z] transforming with rot(a, b, c)
    into a vector transforming with irr_repr(1, a, b, c)
    see assert for usage
    """
    with torch_default_dtype(torch.float64):
        A = torch.tensor([[0, 1, 0], [0, 0, 1], [1, 0, 0]], dtype=torch.float64)
        if check:
            assert all(torch.allclose(irr_repr(1, a, b, c) @ A, A @ rot(a, b, c)) for a, b, c in torch.rand(10, 3))
    return A.type(torch.get_default_dtype())


def irreducible_basis_to_xyz(check=True):
    """
    to convert a vector transforming with irr_repr(1, a, b, c)
    into a vector [x, y, z] transforming with rot(a, b, c)
    see assert for usage

    Inverse of xyz_to_irreducible_basis
    """
    with torch_default_dtype(torch.float64):
        A = torch.tensor([[0, 0, 1], [1, 0, 0], [0, 1, 0]], dtype=torch.float64)
        if check:
            assert all(torch.allclose(A @ irr_repr(1, a, b, c), rot(a, b, c) @ A) for a, b, c in torch.rand(10, 3))
    return A.type(torch.get_default_dtype())


def xyz3x3_repr(a, b, c):
    """
    representation of 3x3 tensors
    T --> R T R^t
    """
    r = rot(a, b, c)
    return kron(r, r)


def xyz3x3_to_irreducible_basis():
    """
    to convert a 3x3 tensor transforming with xyz3x3_repr(a, b, c)
    into its 1 + 3 + 5 component transforming with irr_repr(0, a, b, c), irr_repr(1, a, b, c), irr_repr(3, a, b, c)
    see assert for usage
    """
    with torch_default_dtype(torch.float64):
        to1 = torch.tensor([
            [1, 0, 0, 0, 1, 0, 0, 0, 1],
        ], dtype=torch.get_default_dtype())
        assert all(torch.allclose(irr_repr(0, a, b, c) @ to1, to1 @ xyz3x3_repr(a, b, c)) for a, b, c in torch.rand(10, 3))

        to3 = torch.tensor([
            [0, 0, -1, 0, 0, 0, 1, 0, 0],
            [0, 1, 0, -1, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 1, 0, -1, 0],
        ], dtype=torch.get_default_dtype())
        assert all(torch.allclose(irr_repr(1, a, b, c) @ to3, to3 @ xyz3x3_repr(a, b, c)) for a, b, c in torch.rand(10, 3))

        to5 = torch.tensor([
            [0, 1, 0, 1, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 1, 0, 1, 0],
            [-3**.5 / 3, 0, 0, 0, -3**.5 / 3, 0, 0, 0, 12**.5 / 3],
            [0, 0, 1, 0, 0, 0, 1, 0, 0],
            [1, 0, 0, 0, -1, 0, 0, 0, 0]
        ], dtype=torch.get_default_dtype())
        assert all(torch.allclose(irr_repr(2, a, b, c) @ to5, to5 @ xyz3x3_repr(a, b, c)) for a, b, c in torch.rand(10, 3))

    return to1.type(torch.get_default_dtype()), to3.type(torch.get_default_dtype()), to5.type(torch.get_default_dtype())


def intertwiners(D1, D2, eps=1e-9, with_parity=False):
    """
    Compute a basis of the vector space of matrices A such that
    D1(g) A = A D2(g) for all g in O(3)
    """
    e = (0, 0, 0, 0) if with_parity else (0, 0, 0)
    I1 = D1(*e)
    I2 = D2(*e)

    # picking 20 random rotations seems good enough
    rr = [(rand_angles() + (i % 2,)) if with_parity else rand_angles() for i in range(20)]
    xs = [kron(D1(*g), I2) - kron(I1, D2(*g).T) for g in rr]
    xtx = sum(x.T @ x for x in xs)

    res = xtx.symeig(eigenvectors=True)
    null_space = res.eigenvectors.T[res.eigenvalues.abs() < eps]
    null_space = null_space.reshape(null_space.shape[0], I1.shape[0], I2.shape[0])

    # check that it works
    solutions = []
    for A in null_space:
        d = 0
        for _ in range(4):
            if with_parity:
                r = rand_angles()
                p = torch.randint(0, 2, size=()).item()
                g = r + (p,)
            else:
                g = rand_angles()
            d += A @ D2(*g) - D1(*g) @ A
        d /= 4
        if d.abs().max() < eps:
            solutions.append((d.norm(), A))
    solutions = [A for _, A in sorted(solutions, key=lambda x: x[0])]

    return torch.stack(solutions) if len(solutions) > 0 else torch.zeros(0, I1.shape[0], I2.shape[0])


def reduce(D, D_small, eps=1e-9, with_parity=False):
    """
    Given a "big" representation and a "small" representation
    computes how many times the small appears in the big one and return:
    - how many times the "small" appears in the "big"
    - a matrix that block diagonalize the "big" rep.
    - the remaining of the "big" representation
    """
    def change_and_remove(A, oldD, d):
        def newD(*g):
            return (A @ oldD(*g) @ A.T)[d:][:, d:]
        return newD

    e = (0, 0, 0, 0) if with_parity else (0, 0, 0)
    dim = D(*e).shape[0]
    dim_small = D_small(*e).shape[0]

    D_rest = D
    bigA = torch.eye(dim)
    n = 0

    while True:
        A = intertwiners(D_small, D_rest, eps, with_parity) * dim_small**0.5

        # stops if "small" does not appear in "big" anymore
        if A.shape[0] == 0:
            break

        A, expand = orthonormalize(A[0], eps)
        A = torch.cat([A, expand])

        bigA = direct_sum(torch.eye(n * dim_small), A) @ bigA
        n += 1
        D_rest = change_and_remove(bigA, D, n * dim_small)

    if with_parity:
        g = (5.5407, 1.3256, 2.8139, 1)
    else:
        g = (5.5407, 1.3256, 2.8139)
    assert (bigA @ D(*g) @ bigA.T - direct_sum(*[D_small(*g)] * n + [D_rest(*g)])).abs().max() < eps
    return n, bigA, D_rest


@torch.jit.script
def orthonormalize(
        vecs: torch.Tensor,
        eps: float = 1e-9
) -> Tuple[torch.Tensor, torch.Tensor]:  # pragma: no cover
    """
    :param vecs: tensor of shape [n, m] with n <= m
    :return: (base, expand)

    base.shape[1] == m
    expand.shape[1] == m
    base.shape[0] + expand.shape[0] == m

    cat[base, expand] is orthonormal
    """
    assert vecs.dim() == 2
    dim = vecs.shape[1]

    base = []
    for x in vecs:
        for y in base:
            x -= torch.dot(x, y) * y
        if x.norm() > 2 * eps:
            x = x / x.norm()
            x[x.abs() < eps] = x.new_zeros(())
            base += [x]

    expand = []
    for x in torch.eye(dim, device=vecs.device, dtype=vecs.dtype):
        for y in base + expand:
            x -= torch.dot(x, y) * y
        if x.norm() > 2 * eps:
            x /= x.norm()
            x[x.abs() < eps] = x.new_zeros(())
            x *= x[x.nonzero()[0, 0]].sign()
            expand += [x]

    base = torch.stack(base) if len(base) > 0 else vecs.new_zeros(0, dim)
    expand = torch.stack(expand) if len(expand) > 0 else vecs.new_zeros(0, dim)

    return base, expand



# pylint: disable=arguments-differ, no-member, missing-docstring, invalid-name, line-too-long
import math
from functools import partial

import torch


class ConstantRadialModel(torch.nn.Module):
    def __init__(self, d):
        super().__init__()
        self.weight = torch.nn.Parameter(torch.randn(d))

    def forward(self, _radii):
        batch = _radii.size(0)
        return self.weight.reshape(1, -1).expand(batch, -1)


class FiniteElementModel(torch.nn.Module):
    def __init__(self, out_dim, position, basis, Model):
        '''
        :param out_dim: output dimension
        :param position: tensor [i, ...]
        :param basis: scalar function: tensor [a, ...] -> [a]
        :param Model: Class(d1, d2), trainable model: R^d1 -> R^d2
        '''
        super().__init__()
        self.register_buffer('position', position)
        self.basis = basis
        self.f = Model(len(position), out_dim)

    def forward(self, x):
        """
        :param x: tensor [batch, ...]
        :return: tensor [batch, dim]
        """
        diff = x.unsqueeze(1) - self.position.unsqueeze(0)  # [batch, i, ...]
        batch, n, *rest = diff.size()
        x = self.basis(diff.reshape(-1, *rest)).reshape(batch, n)  # [batch, i]
        return self.f(x)


class FC(torch.nn.Module):
    def __init__(self, d1, d2, h, L, act):
        super().__init__()

        weights = []

        hh = d1
        for _ in range(L):
            weights.append(torch.nn.Parameter(torch.randn(h, hh)))
            hh = h

        weights.append(torch.nn.Parameter(torch.randn(d2, hh)))
        self.weights = torch.nn.ParameterList(weights)
        self.act = act

    def forward(self, x):
        L = len(self.weights) - 1

        if L == 0:
            W = self.weights[0]
            h = x.size(1)
            return x @ (W.t() / h ** 0.5)

        for i, W in enumerate(self.weights):
            h = x.size(1)

            if i == 0:
                # note: normalization assumes that the sum of the inputs is 1
                x = self.act(x @ W.t())
            elif i < L:
                x = self.act(x @ (W.t() / h ** 0.5))
            else:
                x = x @ (W.t() / h ** 0.5)

        return x


def FiniteElementFCModel(out_dim, position, basis, h, L, act):
    Model = partial(FC, h=h, L=L, act=act)
    return FiniteElementModel(out_dim, position, basis, Model)


def CosineBasisModel(out_dim, max_radius, number_of_basis, h, L, act):
    r""" The tex for the math reads
    \begin{cases}
        cos^2 (\gamma (d_{ab} - \mu_{k}) \frac{\pi}{2}) & 1 \geq \gamma (d_{ab} - \mu_{k}) \geq -1 \\
        0 & otherwise
    \end{cases}
    """
    radii = torch.linspace(0, max_radius, steps=number_of_basis)
    step = radii[1] - radii[0]

    def basis(x):
        return x.div(step).add(1).relu().sub(2).neg().relu().add(1).mul(math.pi / 2).cos().pow(2)
    return FiniteElementFCModel(out_dim, radii, basis, h, L, act)


def GaussianRadialModel(out_dim, max_radius, number_of_basis, h, L, act, min_radius=0.):
    """exp(-x^2 /spacing)"""
    spacing = (max_radius - min_radius) / (number_of_basis - 1)
    radii = torch.linspace(min_radius, max_radius, number_of_basis)
    sigma = 0.8 * spacing

    def basis(x):
        return x.div(sigma).pow(2).neg().exp().div(1.423085244900308)
    return FiniteElementFCModel(out_dim, radii, basis, h, L, act)


class BesselRadialModel(torch.nn.Module):
    r"""The tex for the math reads
    \begin{cases}
        \sqrt{\frac{2}{c}} \frac{sin(\frac{n \pi}{c} d)}{d} & 0 \leq x \leq max_radius \\
        0 & otherwise
    \end{cases}
    c = max_radius (cutoff), n = number_of_basis, d = distance (in R_{+})"""
    def __init__(self, out_dim, max_radius, number_of_basis, h, L, act, epsilon=1e-8):
        super().__init__()
        n = torch.linspace(1, number_of_basis, number_of_basis).unsqueeze(0)
        n_scaled = n * math.pi / max_radius
        self.register_buffer('n_scaled', n_scaled)
        self.factor = math.sqrt(2 / max_radius)
        self.max_radius = max_radius
        self.epsilon = epsilon

        self.f = FC(number_of_basis, out_dim, h=h, L=L, act=act)

    def basis(self, x):
        assert x.ndim == 1
        x_within_cutoff = ((x >= 0.0) * (x <= self.max_radius)) * x
        x_within_cutoff = x_within_cutoff.unsqueeze(-1)
        return self.factor * torch.sin(self.n_scaled * x_within_cutoff) / (x_within_cutoff + self.epsilon)

    def forward(self, x):
        x = self.basis(x)
        return self.f(x)



# pylint: disable=not-callable, no-member, invalid-name, line-too-long, arguments-differ, abstract-method
"""
Fourier Transform : sphere (grid) <--> spherical tensor (Rs=[(1, l) for l in range(lmax + 1)])

We use the Fast Fourier Transform for specific
"""
import math

import lie_learn.spaces.S3 as S3
import torch
import torch.fft
from e3nn import rsh, o3
from e3nn.util.default_dtype import torch_default_dtype


def s2_grid(res_beta, res_alpha):
    """
    grid on the sphere
    """
    i = torch.arange(res_beta).to(dtype=torch.get_default_dtype())
    betas = (i + 0.5) / res_beta * math.pi

    i = torch.arange(res_alpha).to(dtype=torch.get_default_dtype())
    alphas = i / res_alpha * 2 * math.pi
    return betas, alphas


def spherical_harmonics_s2_grid(lmax, res_beta, res_alpha, _version=1):
    """
    computes the spherical harmonics on the grid on the sphere
    """
    with torch_default_dtype(torch.float64):
        betas, alphas = s2_grid(res_beta, res_alpha)
        sha = rsh.spherical_harmonics_alpha(lmax, alphas)  # [a, m]
        shb = rsh.spherical_harmonics_z(list(range(lmax + 1)), betas.cos(), betas.sin().abs())  # [b, l * m]
        return betas, alphas, shb, sha


def complete_lmax_res(lmax, res_beta, res_alpha):
    """
    try to use FFT
    i.e. 2 * lmax + 1 == res_alpha
    """
    if res_beta is None:
        res_beta = 2 * (lmax + 1)  # minimum req. to go on sphere and back

    if res_alpha is None:
        if lmax is not None:
            if res_beta is not None:
                res_alpha = max(2 * lmax + 1, res_beta - 1)
            else:
                res_alpha = 2 * lmax + 1  # minimum req. to go on sphere and back
        elif res_beta is not None:
            res_alpha = res_beta - 1

    if lmax is None:
        lmax = min(res_beta // 2 - 1, res_alpha // 2)  # maximum possible to go on sphere and back

    assert res_beta % 2 == 0
    assert lmax + 1 <= res_beta // 2

    return lmax, res_beta, res_alpha


def irfft(x, res):
    """
    ```
    _betas, _alphas, _shb, sha = s2grid.spherical_harmonics_s2_grid(lmax, res, res)
    x = torch.randn(2 * lmax + 1)
    s2grid.irfft(x, res) == sha @ x
    ```
    :param x: tensor of shape [..., m]
    :return: tensor of shape [..., alpha]
    """
    assert res % 2 == 1
    *size, sm = x.shape
    x = x.reshape(-1, sm)
    x = torch.cat([
        x.new_zeros(x.shape[0], (res - sm) // 2),
        x,
        x.new_zeros(x.shape[0], (res - sm) // 2),
    ], dim=-1)
    assert x.shape[1] == res
    l = res // 2
    x = torch.complex(
        torch.cat([
            x[:, l:l + 1],
            x[:, l + 1:].div(math.sqrt(2))
        ], dim=1),
        torch.cat([
            torch.zeros_like(x[:, :1]),
            x[:, :l].flip(-1).div(-math.sqrt(2)),
        ], dim=1),
    )
    x = torch.fft.irfft(x, n=res, dim=1) * res
    return x.reshape(*size, res)


def rfft(x, l):
    """
    ```
    _betas, _alphas, _shb, sha = s2grid.spherical_harmonics_s2_grid(lmax, res, res)
    x = torch.randn(res)
    s2grid.rfft(x, lmax) == x @ sha
    ```
    :param x: tensor of shape [..., alpha]
    :return: tensor of shape [..., m]
    """
    *size, res = x.shape
    x = x.reshape(-1, res)
    x = torch.fft.rfft(x, dim=1)
    x = torch.cat([
        x[:, 1:l + 1].imag.flip(1).mul(-math.sqrt(2)),
        x[:, :1].real,
        x[:, 1:l + 1].real.mul(math.sqrt(2)),
    ], dim=1)
    return x.reshape(*size, 2 * l + 1)


class ToS2Grid(torch.nn.Module):
    """
    Transform spherical tensor into signal on the sphere

    The inverse transformation of FromS2Grid
    """

    def __init__(self, lmax=None, res=None, normalization='component'):
        """
        :param lmax: lmax of the input signal
        :param res: resolution of the output as a tuple (beta resolution, alpha resolution)
        :param normalization: either 'norm', 'component', 'none' or custom
        """
        super().__init__()

        assert normalization in ['norm', 'component', 'none'] or torch.is_tensor(normalization), "normalization needs to be 'norm', 'component' or 'none'"

        if isinstance(res, int) or res is None:
            lmax, res_beta, res_alpha = complete_lmax_res(lmax, res, None)
        else:
            lmax, res_beta, res_alpha = complete_lmax_res(lmax, *res)

        betas, alphas, shb, sha = spherical_harmonics_s2_grid(lmax, res_beta, res_alpha)

        with torch_default_dtype(torch.float64):
            if normalization == 'component':
                # normalize such that all l has the same variance on the sphere
                # given that all componant has mean 0 and variance 1
                n = math.sqrt(4 * math.pi) * torch.tensor([
                    1 / math.sqrt(2 * l + 1)
                    for l in range(lmax + 1)
                ]) / math.sqrt(lmax + 1)
            if normalization == 'norm':
                # normalize such that all l has the same variance on the sphere
                # given that all componant has mean 0 and variance 1/(2L+1)
                n = math.sqrt(4 * math.pi) * torch.ones(lmax + 1) / math.sqrt(lmax + 1)
            if normalization == 'none':
                n = torch.ones(lmax + 1)
            if torch.is_tensor(normalization):
                n = normalization.to(dtype=torch.float64)
            m = rsh.spherical_harmonics_expand_matrix(range(lmax + 1))  # [l, m, i]
        shb = torch.einsum('lmj,bj,lmi,l->mbi', m, shb, m, n)  # [m, b, i]

        self.register_buffer('alphas', alphas)
        self.register_buffer('betas', betas)
        self.register_buffer('sha', sha)
        self.register_buffer('shb', shb)
        self.to(torch.get_default_dtype())

    @property
    def grid(self):
        """
        grid of positions
        """
        beta, alpha = torch.meshgrid(self.betas, self.alphas)
        return o3.angles_to_xyz(alpha, beta)

    def forward(self, x):
        """
        :param x: tensor [..., i=l * m]
        :return: tensor [..., beta, alpha]
        """
        size = x.shape[:-1]
        lmax = round(x.shape[-1] ** 0.5) - 1
        x = x.reshape(-1, (lmax + 1) ** 2)

        x = torch.einsum('mbi,zi->zbm', self.shb, x)  # [batch, beta, m]

        sa, sm = self.sha.shape
        if sa >= sm and sa % 2 == 1:
            x = irfft(x, sa)
        else:
            x = torch.einsum('am,zbm->zba', self.sha, x)
        return x.reshape(*size, *x.shape[1:])


class FromS2Grid(torch.nn.Module):
    """
    Transform signal on the sphere into spherical tensor

    The inverse transformation of ToS2Grid
    """

    def __init__(self, res=None, lmax=None, normalization='component', lmax_in=None):
        """
        :param res: resolution of the input as a tuple (beta resolution, alpha resolution)
        :param lmax: maximum l of the output
        :param normalization: either 'norm', 'component', 'none' or custom
        :param lmax_in: maximum l of the input of ToS2Grid in order to be the inverse
        """
        super().__init__()

        assert normalization in ['norm', 'component', 'none'] or torch.is_tensor(normalization), "normalization needs to be 'norm', 'component' or 'none'"

        if isinstance(res, int) or res is None:
            lmax, res_beta, res_alpha = complete_lmax_res(lmax, res, None)
        else:
            lmax, res_beta, res_alpha = complete_lmax_res(lmax, *res)

        if lmax_in is None:
            lmax_in = lmax

        betas, alphas, shb, sha = spherical_harmonics_s2_grid(lmax, res_beta, res_alpha)

        with torch_default_dtype(torch.float64):
            # normalize such that it is the inverse of ToS2Grid
            if normalization == 'component':
                n = math.sqrt(4 * math.pi) * torch.tensor([
                    math.sqrt(2 * l + 1)
                    for l in range(lmax + 1)
                ]) * math.sqrt(lmax_in + 1)
            if normalization == 'norm':
                n = math.sqrt(4 * math.pi) * torch.ones(lmax + 1) * math.sqrt(lmax_in + 1)
            if normalization == 'none':
                n = 4 * math.pi * torch.ones(lmax + 1)
            if torch.is_tensor(normalization):
                n = normalization.to(dtype=torch.float64)
            m = rsh.spherical_harmonics_expand_matrix(range(lmax + 1))  # [l, m, i]
            assert res_beta % 2 == 0
            qw = torch.tensor(S3.quadrature_weights(res_beta // 2)) * res_beta**2 / res_alpha  # [b]
        shb = torch.einsum('lmj,bj,lmi,l,b->mbi', m, shb, m, n, qw)  # [m, b, i]

        self.register_buffer('alphas', alphas)
        self.register_buffer('betas', betas)
        self.register_buffer('sha', sha)
        self.register_buffer('shb', shb)
        self.to(torch.get_default_dtype())

    @property
    def grid(self):
        """
        grid of positions
        """
        beta, alpha = torch.meshgrid(self.betas, self.alphas)
        return o3.angles_to_xyz(alpha, beta)

    def forward(self, x):
        """
        :param x: tensor [..., beta, alpha]
        :return: tensor [..., i=l * m]
        """
        size = x.shape[:-2]
        res_beta, res_alpha = x.shape[-2:]
        x = x.reshape(-1, res_beta, res_alpha)

        sa, sm = self.sha.shape
        if sm <= sa and sa % 2 == 1:
            x = rfft(x, sm // 2)
        else:
            x = torch.einsum('am,zba->zbm', self.sha, x)
        x = torch.einsum('mbi,zbm->zi', self.shb, x)
        return x.reshape(*size, x.shape[1])



# pylint: disable=not-callable, no-member, invalid-name, line-too-long, missing-docstring, arguments-differ
import torch
from e3nn.tensor.irrep_tensor import IrrepTensor
from e3nn import rs, o3


class CartesianTensor():
    def __init__(self, tensor, formula=None):
        if tuple(tensor.shape) != tuple(3 for i in range(tensor.dim())):
            raise ValueError(f"all dimensions of tensor should have shape 3 but tensor has shape {tensor.shape}")
        if formula is None:
            formula = "abcdefghijklmnopqrstuvxyz"[:tensor.dim()]
        self.formula = formula
        self.tensor = tensor

    def to_irrep_transformation(self):
        dim = self.tensor.dim()
        change = o3.kron(*[o3.xyz_to_irreducible_basis()] * dim)
        Rs = [(1, 1, -1)]  # vectors
        old_indices = self.formula.split("=")[0]
        Rs_out, Q = rs.reduce_tensor(self.formula, **{i: Rs for i in old_indices})
        return Rs_out, torch.einsum('ab,bc->ac', Q, change.reshape(3 ** dim, 3 ** dim))

    def to_irrep_tensor(self):
        Rs_out, Q = self.to_irrep_transformation()
        tensor = torch.einsum('ab,b->a', Q, self.tensor.reshape(-1))
        return IrrepTensor(tensor, Rs_out)



# pylint: disable=invalid-name, arguments-differ, missing-docstring, line-too-long, no-member
import torch

from e3nn import o3, rs


class SO3Activation(torch.nn.Module):
    def __init__(self, Rs, act, n):
        '''
        map to a signal on SO3, apply the non linearity point wise and project back
        the signal on SO3 is the regular representation of SO3
        and we can apply a pointwise operation on these representations

        :param Rs: input representation
        :param act: activation function
        :param n: number of point on the sphere (the higher the more accurate)
        '''
        super().__init__()

        Rs = rs.simplify(Rs)
        mul0, _, _ = Rs[0]
        assert all(mul0 * (2 * l + 1) == mul for mul, l, _ in Rs)
        assert [l for _, l, _ in Rs] == list(range(len(Rs)))
        assert all(p == 0 for _, l, p in Rs)

        self.Rs_out = Rs

        x = [o3.rand_rot() for _ in range(n)]
        Z = torch.stack([torch.cat([o3.irr_repr(l, *o3.rot_to_abc(R)).flatten() * (2 * l + 1)**0.5 for l in range(len(Rs))]) for R in x])  # [z, lmn]
        Z.div_(Z.shape[1]**0.5)
        self.register_buffer('Z', Z)
        self.act = act

    def forward(self, features, dim=-1):
        '''
        :param features: [..., channels, ...]
        expected normalization: component
        '''
        assert features.shape[dim] == self.Z.shape[1]  # assert mul == 1 for now
        features = features.transpose(0, dim)
        out_features = (self.Z.shape[1] / self.Z.shape[0]) * self.Z.T @ self.act(self.Z @ features.flatten(1))
        out_features = out_features.reshape(*features.shape)
        out_features = out_features.transpose(0, dim)
        return out_features



# pylint: disable=missing-docstring
from .point import GatedConvNetwork, GatedConvParityNetwork, S2ConvNetwork, GatedNetwork, MLNetwork
from .image import ImageS2Network, ImageGatedConvNetwork, ImageGatedConvParityNetwork
from .s2 import S2Network, S2ParityNetwork
from .gate import make_gated_block

__all__ = [
    'GatedNetwork',
    'GatedConvNetwork',
    'GatedConvParityNetwork',
    'S2ConvNetwork',
    'ImageS2Network',
    'ImageGatedConvNetwork',
    'ImageGatedConvParityNetwork',
    'S2Network',
    'S2ParityNetwork',
    'make_gated_block',
    'MLNetwork',
]



# pylint: disable=arguments-differ, no-member, missing-docstring, invalid-name, line-too-long
import math

import torch

from e3nn import rs


class KernelLinear(torch.nn.Module):
    def __init__(self, Rs_in, Rs_out):
        """
        :param Rs_in: list of triplet (multiplicity, representation order, parity)
        :param Rs_out: list of triplet (multiplicity, representation order, parity)
        representation order = nonnegative integer
        parity = 0 (no parity), 1 (even), -1 (odd)
        """
        super().__init__()
        self.Rs_in = rs.simplify(Rs_in)
        self.Rs_out = rs.simplify(Rs_out)

        n_path = 0

        for mul_out, l_out, p_out in self.Rs_out:
            for mul_in, l_in, p_in in self.Rs_in:
                if (l_out, p_out) == (l_in, p_in):
                    # compute the number of degrees of freedom
                    n_path += mul_out * mul_in

        self.weight = torch.nn.Parameter(torch.randn(n_path))

    def forward(self):
        """
        :return: tensor [l_out * mul_out * m_out, l_in * mul_in * m_in]
        """
        kernel = self.weight.new_zeros(rs.dim(self.Rs_out), rs.dim(self.Rs_in))
        begin_w = 0

        begin_out = 0
        for mul_out, l_out, p_out in self.Rs_out:
            s_out = slice(begin_out, begin_out + mul_out * (2 * l_out + 1))
            begin_out += mul_out * (2 * l_out + 1)

            n_path = 0

            begin_in = 0
            for mul_in, l_in, p_in in self.Rs_in:
                s_in = slice(begin_in, begin_in + mul_in * (2 * l_in + 1))
                begin_in += mul_in * (2 * l_in + 1)

                if (l_out, p_out) == (l_in, p_in):
                    weight = self.weight[begin_w: begin_w + mul_out * mul_in].reshape(mul_out, mul_in)  # [mul_out, mul_in]
                    begin_w += mul_out * mul_in

                    eye = torch.eye(2 * l_in + 1, dtype=self.weight.dtype, device=self.weight.device)
                    kernel[s_out, s_in] = torch.einsum('uv,ij->uivj', weight, eye).reshape(mul_out * (2 * l_out + 1), mul_in * (2 * l_in + 1))
                    n_path += mul_in

            if n_path > 0:
                kernel[s_out] /= math.sqrt(n_path)

        return kernel


class Linear(torch.nn.Module):
    def __init__(self, Rs_in, Rs_out, allow_unused_inputs=False, allow_zero_outputs=False):
        """
        :param Rs_in: list of triplet (multiplicity, representation order, parity)
        :param Rs_out: list of triplet (multiplicity, representation order, parity)
        representation order = nonnegative integer
        parity = 0 (no parity), 1 (even), -1 (odd)
        """
        super().__init__()
        self.Rs_in = rs.convention(Rs_in)
        self.Rs_out = rs.convention(Rs_out)
        if not allow_unused_inputs:
            self.check_input()
        if not allow_zero_outputs:
            self.check_output()

        self.kernel = KernelLinear(self.Rs_in, self.Rs_out)

    def __repr__(self):
        return "{name} ({Rs_in} -> {Rs_out})".format(
            name=self.__class__.__name__,
            Rs_in=rs.format_Rs(self.Rs_in),
            Rs_out=rs.format_Rs(self.Rs_out),
        )

    def check_input(self):
        for _, l_in, p_in in self.Rs_in:
            if not any((l_in, p_in) == (l_out, p_out) for _, l_out, p_out in self.Rs_out):
                raise ValueError("warning! the input (l={}, p={}) cannot be used".format(l_in, p_in))

    def check_output(self):
        for _, l_out, p_out in self.Rs_out:
            if not any((l_in, p_in) == (l_out, p_out) for _, l_in, p_in in self.Rs_in):
                raise ValueError("warning! the output (l={}, p={}) cannot be generated".format(l_out, p_out))

    def forward(self, features):
        """
        :param features: tensor [..., channel]
        :return:         tensor [..., channel]
        """
        *size, dim_in = features.shape
        features = features.reshape(-1, dim_in)

        output = torch.einsum('ij,zj->zi', self.kernel(), features)

        return output.reshape(*size, -1)



# pylint: disable=missing-docstring, line-too-long, invalid-name, arguments-differ, no-member, pointless-statement, unbalanced-tuple-unpacking
from functools import partial

import torch

from e3nn import o3, rs


class KernelLinear(torch.nn.Module):
    def __init__(self, Rs_in, Rs_out):
        """
        :param Rs_in: list of triplet (multiplicity, representation order, parity)
        :param Rs_out: list of triplet (multiplicity, representation order, parity)
        representation order = nonnegative integer
        parity = 0 (no parity), 1 (even), -1 (odd)
        """
        super().__init__()

        selection_rule = partial(o3.selection_rule_in_out_sh, lmax=0)
        self.tp = rs.TensorProduct(Rs_in, selection_rule, Rs_out, sorted=False)
        self.weight = torch.nn.Parameter(torch.randn(rs.dim(self.tp.Rs_in2)))

    def forward(self):
        """
        :return: tensor [l_out * mul_out * m_out, l_in * mul_in * m_in]
        """
        return self.tp.right(self.weight)


class Linear(torch.nn.Module):
    def __init__(self, Rs_in, Rs_out, allow_unused_inputs=False, allow_zero_outputs=False):
        """
        :param Rs_in: list of triplet (multiplicity, representation order, parity)
        :param Rs_out: list of triplet (multiplicity, representation order, parity)
        representation order = nonnegative integer
        parity = 0 (no parity), 1 (even), -1 (odd)
        """
        super().__init__()
        self.Rs_in = rs.convention(Rs_in)
        self.Rs_out = rs.convention(Rs_out)
        if not allow_unused_inputs:
            self.check_input()
        if not allow_zero_outputs:
            self.check_output()

        self.kernel = KernelLinear(self.Rs_in, self.Rs_out)

    def __repr__(self):
        return "{name} ({Rs_in} -> {Rs_out})".format(
            name=self.__class__.__name__,
            Rs_in=rs.format_Rs(self.Rs_in),
            Rs_out=rs.format_Rs(self.Rs_out),
        )

    def check_input(self):
        for _, l_in, p_in in self.Rs_in:
            if not any((l_in, p_in) == (l_out, p_out) for _, l_out, p_out in self.Rs_out):
                raise ValueError("warning! the input (l={}, p={}) cannot be used".format(l_in, p_in))

    def check_output(self):
        for _, l_out, p_out in self.Rs_out:
            if not any((l_in, p_in) == (l_out, p_out) for _, l_in, p_in in self.Rs_in):
                raise ValueError("warning! the output (l={}, p={}) cannot be generated".format(l_out, p_out))

    def forward(self, features):
        """
        :param features: tensor [..., channel]
        :return:         tensor [..., channel]
        """
        size = features.shape[:-1]
        features = features.reshape(-1, rs.dim(self.Rs_in))

        output = torch.einsum('ij,zj->zi', self.kernel(), features)

        return output.reshape(*size, rs.dim(self.Rs_out))



# pylint: disable=no-member, arguments-differ, redefined-builtin, missing-docstring, line-too-long, invalid-name, abstract-method
import torch
from e3nn import rs
from e3nn.image.convolution import Convolution
from e3nn.image.filter import LowPassFilter
from e3nn.non_linearities import GatedBlock, GatedBlockParity
from e3nn.non_linearities.rescaled_act import sigmoid, swish, tanh
from e3nn.non_linearities.s2 import S2Activation
from e3nn.tensor_product import LearnableTensorSquare


class ImageGatedConvNetwork(torch.nn.Module):
    def __init__(self, Rs_in, Rs_hidden, Rs_out, lmax, size=5, layers=3):
        super().__init__()

        representations = [Rs_in]
        representations += [Rs_hidden] * layers
        representations += [Rs_out]

        def make_layer(Rs_in, Rs_out):
            act = GatedBlock(Rs_out, swish, sigmoid)
            conv = Convolution(
                Rs_in, act.Rs_in, size, lmax=lmax, fuzzy_pixels=True, padding=size // 2)
            return torch.nn.Sequential(conv, act)

        self.layers = torch.nn.Sequential(*[
            make_layer(Rs_layer_in, Rs_layer_out)
            for Rs_layer_in, Rs_layer_out in zip(representations[:-2], representations[1:-1])
        ] + [
            Convolution(representations[-2], representations[-1], size, lmax=lmax, fuzzy_pixels=True)
        ])

    def forward(self, input):
        """
        :param input: tensor of shape [batch, x, y, z, channel]
        """
        return self.layers(input)


class ImageGatedConvParityNetwork(torch.nn.Module):
    def __init__(self, Rs_in, mul, Rs_out, lmax, size=5, layers=3):
        super().__init__()

        modules = []

        Rs = rs.convention(Rs_in)
        for _ in range(layers):
            scalars = [(mul, l, p) for mul, l, p in [(mul, 0, +1),
                                                     (mul, 0, -1)] if rs.haslinearpath(Rs, l, p)]
            act_scalars = [(mul, swish if p == 1 else tanh)
                           for mul, l, p in scalars]

            nonscalars = [(mul, l, p) for l in range(1, lmax + 1)
                          for p in [+1, -1] if rs.haslinearpath(Rs, l, p)]
            gates = [(rs.mul_dim(nonscalars), 0, +1)]
            if rs.haslinearpath(Rs, 0, +1):
                gates = [(rs.mul_dim(nonscalars), 0, +1)]
                act_gates = [(-1, sigmoid)]
            else:
                gates = [(rs.mul_dim(nonscalars), 0, -1)]
                act_gates = [(-1, tanh)]

            act = GatedBlockParity(scalars, act_scalars, gates, act_gates, nonscalars)
            conv = Convolution(Rs, act.Rs_in, size, lmax=lmax, fuzzy_pixels=True, padding=size // 2)

            Rs = act.Rs_out

            block = torch.nn.Sequential(conv, act)
            modules.append(block)

        modules += [Convolution(Rs, Rs_out, size, lmax=lmax, fuzzy_pixels=True, padding=size // 2, allow_unused_inputs=True)]
        self.layers = torch.nn.Sequential(*modules)

    def forward(self, input):
        """
        :param input: tensor of shape [batch, x, y, z, channel]
        """
        return self.layers(input)


class ImageS2Network(torch.nn.Module):
    def __init__(self, Rs_in, mul, lmax, Rs_out, size=5, layers=3):
        super().__init__()

        Rs = rs.simplify(Rs_in)
        Rs_out = rs.simplify(Rs_out)
        Rs_act = list(range(lmax + 1))

        self.mul = mul
        self.layers = []

        for _ in range(layers):
            conv = Convolution(
                Rs, mul * Rs_act, size, lmax=lmax, fuzzy_pixels=True, padding=size // 2)

            # s2 nonlinearity
            act = S2Activation(Rs_act, swish, res=60)
            Rs = mul * act.Rs_out

            pool = LowPassFilter(scale=2.0, stride=2)

            self.layers += [torch.nn.ModuleList([conv, act, pool])]

        self.layers = torch.nn.ModuleList(self.layers)
        self.tail = LearnableTensorSquare(Rs, Rs_out)

    def forward(self, x):
        """
        :param x: [batch, x, y, z, channel_in]
        :return: [batch, x, y, z, channel_out]
        """
        for conv, act, pool in self.layers:
            x = conv(x)

            # put multiplicity into batch
            x = x.reshape(*x.shape[:-1], self.mul, rs.dim(act.Rs_in))
            x = act(x)
            # put back into representation
            x = x.reshape(*x.shape[:-2], self.mul * rs.dim(act.Rs_out))

            x = pool(x)

        x = self.tail(x)
        return x



# pylint: disable=invalid-name, arguments-differ, missing-docstring, line-too-long, no-member
import torch

from e3nn import o3, rs, s2grid


class S2Activation(torch.nn.Module):
    def __init__(self, Rs, act, res, normalization='component', lmax_out=None, random_rot=False):
        '''
        map to the sphere, apply the non linearity point wise and project back
        the signal on the sphere is a quasiregular representation of O3
        and we can apply a pointwise operation on these representations

        :param Rs: input representation of the form [(1, l, p_val * (p_arg)^l) for l in [0, ..., lmax]]
        :param act: activation function
        :param res: resolution of the grid on the sphere (the higher the more accurate)
        :param normalization: either 'norm' or 'component'
        :param lmax_out: maximum l of the output
        :param random_rot: rotate randomly the grid
        '''
        super().__init__()

        Rs = rs.simplify(Rs)
        _, _, p_val = Rs[0]
        _, lmax, _ = Rs[-1]
        assert all(mul == 1 for mul, _, _ in Rs)
        assert [l for _, l, _ in Rs] == [l for l in range(lmax + 1)]
        if all(p == p_val for _, l, p in Rs):
            p_arg = 1
        elif all(p == p_val * (-1) ** l for _, l, p in Rs):
            p_arg = -1
        else:
            assert False, "the parity of the input is not well defined"
        self.Rs_in = Rs
        # the input transforms as : A_l ---> p_val * (p_arg)^l * A_l
        # the sphere signal transforms as : f(r) ---> p_val * f(p_arg * r)
        if lmax_out is None:
            lmax_out = lmax

        if p_val == +1 or p_val == 0:
            self.Rs_out = [(1, l, p_val * p_arg ** l) for l in range(lmax_out + 1)]
        if p_val == -1:
            x = torch.linspace(0, 10, 256)
            a1, a2 = act(x), act(-x)
            if (a1 - a2).abs().max() < a1.abs().max() * 1e-10:
                # p_act = 1
                self.Rs_out = [(1, l, p_arg ** l) for l in range(lmax_out + 1)]
            elif (a1 + a2).abs().max() < a1.abs().max() * 1e-10:
                # p_act = -1
                self.Rs_out = [(1, l, -p_arg ** l) for l in range(lmax_out + 1)]
            else:
                # p_act = 0
                raise ValueError("warning! the parity is violated")

        self.to_s2 = s2grid.ToS2Grid(lmax, res, normalization=normalization)
        self.from_s2 = s2grid.FromS2Grid(res, lmax_out, normalization=normalization, lmax_in=lmax)
        self.act = act
        self.random_rot = random_rot

    def __repr__(self):
        return "{name} ({Rs_in} -> {Rs_out})".format(
            name=self.__class__.__name__,
            Rs_in=rs.format_Rs(self.Rs_in),
            Rs_out=rs.format_Rs(self.Rs_out),
        )

    def forward(self, features):
        '''
        :param features: [..., l * m]
        '''
        assert features.shape[-1] == rs.dim(self.Rs_in)

        if self.random_rot:
            abc = o3.rand_angles()
            features = torch.einsum('ij,...j->...i', rs.rep(self.Rs_in, *abc), features)

        features = self.to_s2(features)  # [..., beta, alpha]
        features = self.act(features)
        features = self.from_s2(features)

        if self.random_rot:
            features = torch.einsum('ij,...j->...i', rs.rep(self.Rs_out, *abc).T, features)
        return features



# pylint: disable=C,R,E1101
import torch
import torch.nn as nn


class GroupNorm(nn.Module):
    def __init__(self, Rs, eps=1e-5, affine=True):
        '''
        :param Rs: list of tuple (multiplicity, dimension)
        '''
        super().__init__()

        self.Rs = [(m, d) for m, d in Rs if m * d > 0]

        self.eps = eps
        self.affine = affine

        if affine:
            self.weight = nn.Parameter(torch.ones(sum([m for m, d in Rs])))
            self.bias = nn.Parameter(torch.zeros(sum([m for m, d in Rs if d == 1])))
        else:
            self.register_parameter('weight', None)
            self.register_parameter('bias', None)

    def __repr__(self):
        return "{name} (Rs={Rs}, eps={eps}, affine={affine})".format(name=self.__class__.__name__, **self.__dict__)

    def forward(self, input):  # pylint: disable=W
        '''
        :param input: [batch, stacked feature, x, y, z]
        '''

        fields = []
        ix = 0
        iw = 0
        ib = 0
        for m, d in self.Rs:
            field = input[:, ix: ix + m * d]  # [batch, feature * repr, x, y, z]
            ix += m * d
            field = field.reshape(input.size(0), m, d, -1)  # [batch, feature, repr, x * y * z]

            if d == 1:  # scalars
                field_mean = field.reshape(input.size(0), -1).mean(-1)  # [batch]
                field = field - field_mean.reshape(-1, 1, 1, 1)  # [batch, feature, repr, x * y * z]

            field_norm = torch.sum(field ** 2, dim=2)  # [batch, feature, x * y * z]
            field_norm = field_norm.reshape(input.size(0), -1).mean(-1)  # [batch]
            field_norm = (field_norm + self.eps).pow(-0.5).reshape(-1, 1, 1, 1)  # [batch, feature, repr, x * y * z]

            if self.affine:
                weight = self.weight[iw: iw + m]  # [feature]
                iw += m
                field_norm = field_norm * weight.reshape(1, -1, 1, 1)  # [batch, feature, repr, x * y * z]

            field = field * field_norm  # [batch, feature, repr, x * y * z]

            if self.affine and d == 1:  # scalars
                bias = self.bias[ib: ib + m]  # [feature]
                ib += m
                field = field + bias.reshape(1, -1, 1, 1)  # [batch, feature, repr, x * y * z]

            fields.append(field.reshape(input.size(0), m * d, *input.size()[2:]))

        assert ix == input.size(1)
        if self.affine:
            assert iw == self.weight.size(0)
            assert ib == self.bias.size(0)
        return torch.cat(fields, dim=1)  # [batch, stacked feature, x, y, z]



# pylint: disable=no-member, missing-docstring, invalid-name, redefined-builtin, arguments-differ, line-too-long, unbalanced-tuple-unpacking
from time import perf_counter
import torch

DATA_TIMES = {}


def clear():
    global DATA_TIMES
    DATA_TIMES = {}


def start():
    if torch.cuda.is_available():
        torch.cuda.synchronize()
    return perf_counter()


def end(name, begin_time):
    global DATA_TIMES

    if torch.cuda.is_available():
        torch.cuda.synchronize()
    end_time = perf_counter()
    delta = end_time - begin_time

    try:
        DATA_TIMES[name].append(delta)
    except KeyError:
        DATA_TIMES[name] = [delta]
    return end_time


def text_statistics():
    text = "[time logging] ...............unit is seconds... [tot time]/ [nbr] = [per call] [percent]\n"
    total = max(sum(times) for _, times in DATA_TIMES.items())

    for name, times in sorted(DATA_TIMES.items(), key=lambda x: sum(x[1]), reverse=True):
        text += "[time logging] {:.<30}... {: >9.3} / {: <5} = {: <10.3} {}%\n".format(
            name, sum(times),
            len(times), sum(times) / len(times),
            round(100 * sum(times) / total))
    return text



# pylint: disable=no-member, arguments-differ, redefined-builtin, missing-docstring, line-too-long, invalid-name, abstract-method
from functools import partial

import torch

from e3nn import o3, rs
from e3nn.kernel import Kernel
from e3nn.non_linearities import GatedBlock, GatedBlockParity
from e3nn.non_linearities.rescaled_act import sigmoid, swish, tanh
from e3nn.non_linearities.s2 import S2Activation
from e3nn.point.operations import Convolution
from e3nn.radial import GaussianRadialModel
from e3nn.util.deprecation import deprecated
from e3nn.networks.gate import make_gated_block


def GatedNetwork(Rs_in, Rs_out, mul, lmax, layers=3, min_radius=0.0, max_radius=1.0, radial_basis=3, radial_neurons=100, radial_layers=2, radial_act=swish):
    def layer(Rs1, Rs2):
        R = partial(
            GaussianRadialModel,
            max_radius=max_radius,
            number_of_basis=radial_basis,
            h=radial_neurons,
            L=radial_layers,
            act=radial_act,
            min_radius=min_radius
        )
        k = Kernel(Rs1, Rs2, R, partial(o3.selection_rule_in_out_sh, lmax=lmax), allow_unused_inputs=True)
        return Convolution(k)
    def activation(Rs):
        return make_gated_block(Rs, mul, lmax)
    return MLNetwork(Rs_in, Rs_out, layer, activation, layers)


class MLNetwork(torch.nn.Module):
    def __init__(self, Rs_in, Rs_out, layer, activation, layers=3):
        super().__init__()

        modules = []

        Rs = rs.convention(Rs_in)
        for _ in range(layers):
            act = activation(Rs)
            lay = layer(Rs, act.Rs_in)

            Rs = act.Rs_out

            modules += [torch.nn.ModuleList([lay, act])]

        self.layers = torch.nn.ModuleList(modules)
        self.layers.append(layer(Rs, Rs_out))

    def forward(self, x, *args, **kwargs):
        for lay, act in self.layers[:-1]:
            x = lay(x, *args, **kwargs)
            x = act(x)

        return self.layers[-1](x, *args, **kwargs)


@deprecated
class GatedConvNetwork(torch.nn.Module):
    """A basic network architecture that alternates between Convolutions and GatedBlock nonlinearities.

    This network architecture should be used if you only want to be equivariant to rotations and translations but not inversion. GatedBlock acts on irreps of SO(3), meaning it does not include parity. If you want mirror symmetry and inversion, use GatedConvParityNetwork which handles the full irreps of O(3).

    Args:
        Rs_in (rs.TY_RS_STRICT): Representation list of input data. Parity must be omitted or set to 0.
        Rs_hidden (rs.TY_RS_STRICT): Representation list of intermediate data. Parity must be omitted or set to 0.
        Rs_out (rs.TY_RS_STRICT): Representation list of output data. Parity must be omitted or set to 0.
        lmax (int > 0): Maximum L used for spherical harmonic in kernel.
        layers (int, optional): Number of convolution + GatedBlock layers. Defaults to 3.
        max_radius (float, optional): Maximum radius of radial basis functions. Used to initialize RadialFunction for kernel. Defaults to 1.0.
        number_of_basis (int, optional): Number of spaced Gaussian radial basis functions used for RadialFunction. Defaults to 3.
        radial_layers (int, optional): Number of dense layers applied to radial basis functions to create RadialFunction. Defaults to 3.
        kernel (kernel class, optional): Equivariant kernel class. Defaults to e3nn.kernel.Kernel.
        convolution (convolution class, optional): Equivariant convoltion operation. Defaults to e3nn.point.operations.Convolution. For torch_geometric.data.Data input use e3nn.point.message_passing.Convolution.
        min_radius (float, optional): Minimum radius of radial basis functions. Used to initialize RadialFunction for kernel. Defaults to 0.0.

    Example::

        ### For use on torch_geometric.data.Data objects. Otherwise, do not replace convolution kwarg.

        N_atom_types = 3  # For example hydrogen, carbon, oxygen
        Rs_in = [(N_atom_types, 0, 0)]  # GatedBlock acts on irreps of SO(3) so parity (last number) should be omitted or 0.
        Rs_hidden = [(6, 0, 0), (4, 1, 0), (2, 2, 0)]  # 6 L=0, 4 L=1, and 2 L=2 features.
        Rs_out = [(1, 1, 0)]  # Predict vectors on each atom.

        model_kwargs = {
            'convolution': e3nn.point.message_passing.Convolution,  # Use this convolution with data from e3nn.data.DataNeighbors or e3nn.data.DataPeriodicNeighbors
            'Rs_in': Rs_in,
            'Rs_hidden': Rs_hidden,
            'Rs_out': Rs_out,
            'lmax': 2,  # Includes spherical harmonics 0, 1, and 2
            'layers': 3,  # 3 * (conv + nonlin) + conv layers
            'max_radius': r_max,  # Should match whatever is used to generate edges with e3nn.data.DataNeighbors or e3nn.data.DataPeriodicNeighbors
            'number_of_basis': 10,
        }

        model = GatedConvNetwork(**model_kwargs)
    """

    def __init__(self, Rs_in, Rs_hidden, Rs_out, lmax=3, layers=3,
                 max_radius=1.0, number_of_basis=3, radial_layers=3,
                 kernel=Kernel, convolution=Convolution,
                 min_radius=0.0):
        super().__init__()

        representations = [Rs_in]
        representations += [Rs_hidden] * layers
        representations += [Rs_out]

        RadialModel = partial(GaussianRadialModel, max_radius=max_radius,
                              min_radius=min_radius,
                              number_of_basis=number_of_basis, h=100,
                              L=radial_layers, act=swish)

        K = partial(kernel, RadialModel=RadialModel, selection_rule=partial(
            o3.selection_rule_in_out_sh, lmax=lmax))

        def make_layer(Rs_in, Rs_out):
            act = GatedBlock(Rs_out, swish, sigmoid)
            conv = convolution(K(Rs_in, act.Rs_in))
            return torch.nn.ModuleList([conv, act])

        self.layers = torch.nn.ModuleList([
            make_layer(Rs_layer_in, Rs_layer_out)
            for Rs_layer_in, Rs_layer_out in zip(representations[:-2], representations[1:-1])
        ])

        self.layers.append(convolution(
            K(representations[-2], representations[-1])))

    def forward(self, input, *args, **kwargs):
        """Consult the convolution operation used to initalize this class for specifics on what input should be given.

        Args:
            input (torch.tensor with dtype=torch.float64): input tensor with shape [batch, N, C] if default or [N, C] if e3nn.point.message_passing.Convolution

        Examples::

            ## Example for e3nn.point.message_passing.Convolution

            n_norm = 6  # Normalization of convolution to keep variance of output close to 1. Typically, this is equal to slightly larger than the dataset average for number of nearest neighbors.
            data = dh.DataNeighbors(features, Rs_in, pos, r_max, y=target)
            # data.x is input with shape [num_nodes, channels], data.edge_index has shape [2, num_edges], data.edge_attr stores relative distance vectors and has shape [num_edges, 3 (xyz)]
            output = model(data.x, data.edge_index, data.edge_attr, n_norm=n_norm)
            loss = ((output - target) ** 2).mean()
        """
        output = input
        N = args[0].shape[-2]
        if 'n_norm' not in kwargs:
            kwargs['n_norm'] = N

        for conv, act in self.layers[:-1]:
            output = conv(output, *args, **kwargs)
            output = act(output)

        layer = self.layers[-1]
        output = layer(output, *args, **kwargs)
        return output


@deprecated
class GatedConvParityNetwork(torch.nn.Module):
    """A basic network architecture that alternates between Convolutions and GatedBlockParity nonlinearities.

    This network architecture should be used if you want to be fully equivariant to rotations, translations, and inversion (including mirrors).

    Args:
        Rs_in (rs.TY_RS_STRICT): Representation list of input data. Must have parity specified (1 for even, -1 for odd).
        mul (int > 0): multiplicity of irreps of intermediate data.
        Rs_out (rs.TY_RS_STRICT): Representation list of output data. Must have parity specified (1 for even, -1 for odd).
        lmax (int > 0): Maximum L used for spherical harmonic in kernel and sets max L for intermediate data.
        layers (int, optional): Number of convolution + GatedBlock layers. Defaults to 3.
        max_radius (float, optional): Maximum radius of radial basis functions. Used to initialize RadialFunction for kernel. Defaults to 1.0.
        number_of_basis (int, optional): Number of spaced Gaussian radial basis functions used for RadialFunction. Defaults to 3.
        radial_layers (int, optional): Number of dense layers applied to radial basis functions to create RadialFunction. Defaults to 3.
        kernel (kernel class, optional): Equivariant kernel class. Defaults to e3nn.kernel.Kernel.
        convolution (convolution class, optional): Equivariant convolution operation. Defaults to e3nn.point.operations.Convolution. For torch_geometric.data.Data input use e3nn.point.message_passing.Convolution.
        min_radius (float, optional): Minimum radius of radial basis functions. Used to initialize RadialFunction for kernel. Defaults to 0.0.

    Example::

        ### For use on torch_geometric.data.Data objects. Otherwise, do not replace convolution kwarg.

        N_atom_types = 3  # For example hydrogen, carbon, oxygen
        Rs_in = [(N_atom_types, 0, 0)]  # GatedBlock acts on irreps of SO(3) so parity (last number) should be omitted or 0.
        Rs_out = [(1, 1, 0)]  # Predict vectors on each atom.

        model_kwargs = {
            'convolution': e3nn.point.message_passing.Convolution,  # Use this convolution with data from e3nn.data.DataNeighbors or e3nn.data.DataPeriodicNeighbors
            'Rs_in': Rs_in,
            'Rs_out': Rs_out,
            'mul': 4,  # An example intermediate Rs might be [(4, 0, 1), (4, 0, -1), (4, 1, 1), (4, 1, -1), (4, 2, 1), (4, 2, -1)]
            'lmax': 2,  # Includes spherical harmonics 0, 1, and 2 in kernel and intermediate tensors go up to L=2.
            'layers': 3,  # 3 * (conv + nonlin) + conv layers
            'max_radius': r_max,  # Should match whatever is used to generate edges with e3nn.data.DataNeighbors or e3nn.data.DataPeriodicNeighbors
            'number_of_basis': 10,
        }

        model = GatedConvParityNetwork(**model_kwargs)
    """

    def __init__(self, Rs_in, mul, Rs_out, lmax, layers=3,
                 max_radius=1.0, number_of_basis=3, radial_layers=3,
                 kernel=Kernel, convolution=Convolution,
                 min_radius=0.0):
        super().__init__()

        R = partial(GaussianRadialModel, max_radius=max_radius,
                    number_of_basis=number_of_basis, h=100,
                    L=radial_layers, act=swish, min_radius=min_radius)
        K = partial(kernel, RadialModel=R, selection_rule=partial(
            o3.selection_rule_in_out_sh, lmax=lmax))

        modules = []

        Rs = rs.convention(Rs_in)
        for _ in range(layers):
            scalars = [(mul, l, p) for mul, l, p in [(mul, 0, +1),
                                                     (mul, 0, -1)] if rs.haslinearpath(Rs, l, p)]
            act_scalars = [(mul, swish if p == 1 else tanh)
                           for mul, l, p in scalars]

            nonscalars = [(mul, l, p) for l in range(1, lmax + 1)
                          for p in [+1, -1] if rs.haslinearpath(Rs, l, p)]
            if rs.haslinearpath(Rs, 0, +1):
                gates = [(rs.mul_dim(nonscalars), 0, +1)]
                act_gates = [(-1, sigmoid)]
            else:
                gates = [(rs.mul_dim(nonscalars), 0, -1)]
                act_gates = [(-1, tanh)]

            act = GatedBlockParity(scalars, act_scalars,
                                   gates, act_gates, nonscalars)
            conv = convolution(K(Rs, act.Rs_in))

            Rs = act.Rs_out

            block = torch.nn.ModuleList([conv, act])
            modules.append(block)

        self.layers = torch.nn.ModuleList(modules)

        K = partial(K, allow_unused_inputs=True)
        self.layers.append(convolution(K(Rs, Rs_out)))

    def forward(self, input, *args, **kwargs):
        """Consult the convolution operation used to initalize this class for specifics on what input should be given.

        Args:
            input (torch.tensor with dtype=torch.float64): input tensor with shape [batch, N, C] if default or [N, C] if e3nn.point.message_passing.Convolution

        Examples::

            ## Example for e3nn.point.message_passing.Convolution

            n_norm = 6  # Normalization of convolution to keep variance of output close to 1. Typically, this is equal to slightly larger than the dataset average for number of nearest neighbors.
            data = dh.DataNeighbors(features, Rs_in, pos, r_max, y=target)
            # data.x is input with shape [num_nodes, channels], data.edge_index has shape [2, num_edges], data.edge_attr stores relative distance vectors and has shape [num_edges, 3 (xyz)]
            output = model(data.x, data.edge_index, data.edge_attr, n_norm=n_norm)
            loss = ((output - target) ** 2).mean()
        """
        output = input
        N = args[0].shape[-2]
        if 'n_norm' not in kwargs:
            kwargs['n_norm'] = N

        for conv, act in self.layers[:-1]:
            output = conv(output, *args, **kwargs)
            output = act(output)

        layer = self.layers[-1]
        output = layer(output, *args, **kwargs)
        return output


class S2ConvNetwork(torch.nn.Module):
    def __init__(self, Rs_in, mul, Rs_out, lmax, layers=3,
                 max_radius=1.0, number_of_basis=3, radial_layers=3,
                 kernel=Kernel, convolution=Convolution):
        super().__init__()

        Rs_hidden = [(1, l, (-1)**l) for i in range(mul)
                     for l in range(lmax + 1)]
        representations = [Rs_in]
        representations += [Rs_hidden] * layers
        representations += [Rs_out]

        RadialModel = partial(GaussianRadialModel, max_radius=max_radius,
                              number_of_basis=number_of_basis, h=100,
                              L=radial_layers, act=swish)

        K = partial(kernel, RadialModel=RadialModel, selection_rule=partial(
            o3.selection_rule_in_out_sh, lmax=lmax))

        def make_layer(Rs_in, Rs_out):
            act = S2Activation([(1, l, (-1)**l) for l in range(lmax + 1)],
                               sigmoid, lmax_out=lmax, res=20 * (lmax + 1))
            conv = convolution(K(Rs_in, Rs_out))
            return torch.nn.ModuleList([conv, act])

        self.layers = torch.nn.ModuleList([
            make_layer(Rs_layer_in, Rs_layer_out)
            for Rs_layer_in, Rs_layer_out in zip(representations[:-2], representations[1:-1])
        ])

        self.layers.append(convolution(
            K(representations[-2], representations[-1])))
        self.mul = mul
        self.lmax = lmax

    def forward(self, input, *args, **kwargs):
        output = input
        N = args[0].shape[-2]
        if 'n_norm' not in kwargs:
            kwargs['n_norm'] = N

        for conv, act in self.layers[:-1]:
            output = conv(output, *args, **kwargs)
            shape = list(output.shape)
            # Split multiplicities into new batch
            output = output.reshape(shape[:-1] + [self.mul, (self.lmax + 1) ** 2])
            output = act(output)
            output = output.reshape(shape)

        layer = self.layers[-1]
        output = layer(output, *args, **kwargs)
        return output



# pylint: disable=missing-docstring, line-too-long, invalid-name, arguments-differ, no-member, pointless-statement, abstract-method
import math

import torch

from e3nn import o3, rs, rsh
from e3nn.linear import KernelLinear


class Kernel(torch.nn.Module):
    def __init__(self, Rs_in, Rs_out, RadialModel,
                 selection_rule=o3.selection_rule_in_out_sh,
                 normalization='component',
                 allow_unused_inputs=False,
                 allow_zero_outputs=False):
        """
        :param Rs_in: list of triplet (multiplicity, representation order, parity)
        :param Rs_out: list of triplet (multiplicity, representation order, parity)
        :param RadialModel: Class(d), trainable model: R -> R^d
        :param selection_rule: function of signature (l_in, p_in, l_out, p_out) -> [l_filter]
        :param normalization: either 'norm' or 'component'
        representation order = nonnegative integer
        parity = 0 (no parity), 1 (even), -1 (odd)
        """
        super().__init__()
        self.Rs_in = rs.simplify(Rs_in)
        self.Rs_out = rs.simplify(Rs_out)

        self.selection_rule = selection_rule
        if not allow_unused_inputs:
            self.check_input()
        if not allow_zero_outputs:
            self.check_output()

        assert normalization in ['norm', 'component'], "normalization needs to be 'norm' or 'component'"
        self.normalization = normalization

        def lm_normalization(l_out, l_in):
            # put 2l_in+1 to keep the norm of the m vector constant
            # put 2l_ou+1 to keep the variance of each m component constant
            # sum_m Y_m^2 = (2l+1)/(4pi)  and  norm(Q) = 1  implies that norm(QY) = sqrt(1/4pi)
            lm_norm = None
            if normalization == 'norm':
                lm_norm = math.sqrt(2 * l_in + 1) * math.sqrt(4 * math.pi)
            elif normalization == 'component':
                lm_norm = math.sqrt(2 * l_out + 1) * math.sqrt(4 * math.pi)
            return lm_norm

        norm_coef = torch.zeros((len(self.Rs_out), len(self.Rs_in)))

        n_path = 0
        set_of_l_filters = set()

        for i, (mul_out, l_out, p_out) in enumerate(self.Rs_out):
            # consider that we sum a bunch of [lambda_(m_out)] vectors
            # we need to count how many of them we sum in order to normalize the network
            num_summed_elements = 0
            for mul_in, l_in, p_in in self.Rs_in:
                l_filters = self.selection_rule(l_in, p_in, l_out, p_out)
                num_summed_elements += mul_in * len(l_filters)

            for j, (mul_in, l_in, p_in) in enumerate(self.Rs_in):
                # normalization assuming that each terms are of order 1 and uncorrelated
                norm_coef[i, j] = lm_normalization(l_out, l_in) / math.sqrt(num_summed_elements)

                l_filters = self.selection_rule(l_in, p_in, l_out, p_out)
                assert l_filters == sorted(set(l_filters)), "selection_rule must return a sorted list of unique values"
                if p_out != 0:
                    assert all(p_in * (-1) ** l == p_out for l in l_filters), "selection_rule must return l's compatible with SH parity"

                # compute the number of degrees of freedom
                n_path += mul_out * mul_in * len(l_filters)

                # create the set of all spherical harmonics orders needed
                set_of_l_filters = set_of_l_filters.union(l_filters)

        # create the radial model: R+ -> R^n_path
        # it contains the learned parameters
        self.R = RadialModel(n_path)
        self.set_of_l_filters = sorted(set_of_l_filters)
        self.register_buffer('norm_coef', norm_coef)

        self.linear = KernelLinear(self.Rs_in, self.Rs_out)

    def __repr__(self):
        return "{name} ({Rs_in} -> {Rs_out})".format(
            name=self.__class__.__name__,
            Rs_in=rs.format_Rs(self.Rs_in),
            Rs_out=rs.format_Rs(self.Rs_out),
        )

    def check_input(self):
        for _, l_in, p_in in self.Rs_in:
            if not any(self.selection_rule(l_in, p_in, l_out, p_out) for _, l_out, p_out in self.Rs_out):
                raise ValueError("warning! the input (l={}, p={}) cannot be used".format(l_in, p_in))

    def check_output(self):
        for _, l_out, p_out in self.Rs_out:
            if not any(self.selection_rule(l_in, p_in, l_out, p_out) for _, l_in, p_in in self.Rs_in):
                raise ValueError("warning! the output (l={}, p={}) cannot be generated".format(l_out, p_out))

    def forward(self, r, r_eps=0, custom_backward=False):
        """
        :param r: tensor [..., 3]
        :param custom_backward: call KernelFn rather than using automatic differentiation
        :return: tensor [..., l_out * mul_out * m_out, l_in * mul_in * m_in]
        """
        *size, xyz = r.size()
        assert xyz == 3
        r = r.reshape(-1, 3)

        radii = r.norm(2, dim=1)  # [batch]

        # (1) Case r > 0

        # precompute all needed spherical harmonics
        Y = rsh.spherical_harmonics_xyz(self.set_of_l_filters, r[radii > r_eps])  # [batch, l_filter * m_filter]

        # use the radial model to fix all the degrees of freedom
        # note: for the normalization we assume that the variance of R[i] is one
        R = self.R(radii[radii > r_eps])  # [batch, l_out * l_in * mul_out * mul_in * l_filter]

        if custom_backward:
            kernel1 = KernelFn.apply(Y, R, self.norm_coef, self.Rs_in, self.Rs_out, self.selection_rule, self.set_of_l_filters)
        else:
            kernel1 = kernel_fn_forward(Y, R, self.norm_coef, self.Rs_in, self.Rs_out, self.selection_rule, self.set_of_l_filters)

        # (2) Case r = 0

        kernel2 = self.linear()

        kernel = r.new_zeros(len(r), *kernel2.shape)
        kernel[radii > r_eps] = kernel1
        kernel[radii <= r_eps] = kernel2

        return kernel.reshape(*size, *kernel2.shape)


class GroupKernel(torch.nn.Module):
    def __init__(self, Rs_in, Rs_out, kernel, groups):
        super().__init__()
        self.kernels = [kernel(Rs_in, Rs_out) for i in range(groups)]

    def forward(self, *args, **kwargs):
        return torch.stack([ker(*args, **kwargs) for ker in self.kernels], dim=-3)  # [..., group, cout, cin]


def kernel_fn_forward(Y, R, norm_coef, Rs_in, Rs_out, selection_rule, set_of_l_filters):
    """
    :param Y: tensor [batch, l_filter * m_filter]
    :param R: tensor [batch, l_out * l_in * mul_out * mul_in * l_filter]
    :param norm_coef: tensor [l_out, l_in]
    :return: tensor [batch, l_out * mul_out * m_out, l_in * mul_in * m_in]
    """
    batch = Y.shape[0]
    n_in = rs.dim(Rs_in)
    n_out = rs.dim(Rs_out)

    kernel = Y.new_zeros(batch, n_out, n_in)

    # note: for the normalization we assume that the variance of R[i] is one
    begin_R = 0

    begin_out = 0
    for i, (mul_out, l_out, p_out) in enumerate(Rs_out):
        s_out = slice(begin_out, begin_out + mul_out * (2 * l_out + 1))
        begin_out += mul_out * (2 * l_out + 1)

        begin_in = 0
        for j, (mul_in, l_in, p_in) in enumerate(Rs_in):
            s_in = slice(begin_in, begin_in + mul_in * (2 * l_in + 1))
            begin_in += mul_in * (2 * l_in + 1)

            l_filters = selection_rule(l_in, p_in, l_out, p_out)
            if not l_filters:
                continue

            # extract the subset of the `R` that corresponds to the couple (l_out, l_in)
            n = mul_out * mul_in * len(l_filters)
            sub_R = R[:, begin_R: begin_R + n].reshape(batch, mul_out, mul_in, len(l_filters))  # [batch, mul_out, mul_in, l_filter]
            begin_R += n

            # note: I don't know if we can vectorize this for loop because [l_filter * m_filter] cannot be put into [l_filter, m_filter]
            K = 0
            for k, l_filter in enumerate(l_filters):
                tmp = sum(2 * l + 1 for l in set_of_l_filters if l < l_filter)
                sub_Y = Y[:, tmp: tmp + 2 * l_filter + 1]  # [batch, m]

                C = o3.wigner_3j(l_out, l_in, l_filter, cached=True, like=kernel)  # [m_out, m_in, m]

                # note: The multiplication with `sub_R` could also be done outside of the for loop
                K += norm_coef[i, j] * torch.einsum("ijk,zk,zuv->zuivj", (C, sub_Y, sub_R[..., k]))  # [batch, mul_out, m_out, mul_in, m_in]

            if not isinstance(K, int):
                kernel[:, s_out, s_in] = K.reshape_as(kernel[:, s_out, s_in])
    return kernel


class KernelFn(torch.autograd.Function):
    """
    The math is presented here:
    https://slides.com/mariogeiger/backward/
    """
    @staticmethod
    def forward(ctx, Y, R, norm_coef, Rs_in, Rs_out, selection_rule, set_of_l_filters):
        f"""{kernel_fn_forward.__doc__}"""
        ctx.Rs_in = Rs_in
        ctx.Rs_out = Rs_out
        ctx.selection_rule = selection_rule
        ctx.set_of_l_filters = set_of_l_filters

        # save necessary tensors for backward
        saved_Y = saved_R = None
        if Y.requires_grad:
            ctx.Y_shape = Y.shape
            saved_R = R
        if R.requires_grad:
            ctx.R_shape = R.shape
            saved_Y = Y
        ctx.save_for_backward(saved_Y, saved_R, norm_coef)

        return kernel_fn_forward(Y, R, norm_coef, ctx.Rs_in, ctx.Rs_out, ctx.selection_rule, ctx.set_of_l_filters)

    @staticmethod
    def backward(ctx, grad_kernel):  # pragma: no cover
        Y, R, norm_coef = ctx.saved_tensors

        grad_Y = grad_R = None

        if ctx.needs_input_grad[0]:
            grad_Y = grad_kernel.new_zeros(*ctx.Y_shape)  # [batch, l_filter * m_filter]
        if ctx.needs_input_grad[1]:
            grad_R = grad_kernel.new_zeros(*ctx.R_shape)  # [batch, l_out * l_in * mul_out * mul_in * l_filter]

        begin_R = 0

        begin_out = 0
        for i, (mul_out, l_out, p_out) in enumerate(ctx.Rs_out):
            s_out = slice(begin_out, begin_out + mul_out * (2 * l_out + 1))
            begin_out += mul_out * (2 * l_out + 1)

            begin_in = 0
            for j, (mul_in, l_in, p_in) in enumerate(ctx.Rs_in):
                s_in = slice(begin_in, begin_in + mul_in * (2 * l_in + 1))
                begin_in += mul_in * (2 * l_in + 1)

                l_filters = ctx.selection_rule(l_in, p_in, l_out, p_out)
                if not l_filters:
                    continue

                n = mul_out * mul_in * len(l_filters)
                if grad_Y is not None:
                    sub_R = R[:, begin_R: begin_R + n].reshape(
                        -1, mul_out, mul_in, len(l_filters)
                    )  # [batch, mul_out, mul_in, l_filter]
                if grad_R is not None:
                    sub_grad_R = grad_R[:, begin_R: begin_R + n].reshape(
                        -1, mul_out, mul_in, len(l_filters)
                    )  # [batch, mul_out, mul_in, l_filter]
                begin_R += n

                grad_K = grad_kernel[:, s_out, s_in].reshape(-1, mul_out, 2 * l_out + 1, mul_in, 2 * l_in + 1)

                for k, l_filter in enumerate(l_filters):
                    tmp = sum(2 * l + 1 for l in ctx.set_of_l_filters if l < l_filter)
                    C = o3.wigner_3j(l_out, l_in, l_filter, cached=True, like=grad_kernel)  # [m_out, m_in, m]

                    if grad_Y is not None:
                        grad_Y[:, tmp: tmp + 2 * l_filter + 1] += norm_coef[i, j] * torch.einsum("zuivj,ijk,zuv->zk", grad_K, C, sub_R[..., k])
                    if grad_R is not None:
                        sub_Y = Y[:, tmp: tmp + 2 * l_filter + 1]  # [batch, m]
                        sub_grad_R[..., k] = norm_coef[i, j] * torch.einsum("zuivj,ijk,zk->zuv", grad_K, C, sub_Y)

        del ctx
        return grad_Y, grad_R, None, None, None, None, None



# pylint: disable=missing-docstring, line-too-long, invalid-name, arguments-differ
# pylint: disable=no-member, pointless-statement, unbalanced-tuple-unpacking, abstract-method
import math

import torch

from e3nn import o3, rs, rsh
from e3nn.linear_mod import KernelLinear


class Kernel(torch.nn.Module):
    def __init__(self, Rs_in, Rs_out, RadialModel,
                 selection_rule=o3.selection_rule_in_out_sh,
                 normalization='component',
                 allow_unused_inputs=False,
                 allow_zero_outputs=False):
        """
        :param Rs_in: list of triplet (multiplicity, representation order, parity)
        :param Rs_out: list of triplet (multiplicity, representation order, parity)
        :param RadialModel: Class(d), trainable model: R -> R^d
        :param selection_rule: function of signature (l_in, p_in, l_out, p_out) -> [l_filter]
        :param sh: spherical harmonics function of signature ([l_filter], xyz[..., 3]) -> Y[m, ...]
        :param normalization: either 'norm' or 'component'
        representation order = nonnegative integer
        parity = 0 (no parity), 1 (even), -1 (odd)
        """
        super().__init__()

        self.Rs_in = rs.convention(Rs_in)
        self.Rs_out = rs.convention(Rs_out)
        if not allow_unused_inputs:
            self.check_input(selection_rule)
        if not allow_zero_outputs:
            self.check_output(selection_rule)

        self.normalization = normalization

        self.tp = rs.TensorProduct(self.Rs_in, selection_rule, Rs_out, normalization, sorted=True)
        self.Rs_f = self.tp.Rs_in2

        self.Ls = [l for _, l, _ in self.Rs_f]
        self.R = RadialModel(rs.mul_dim(self.Rs_f))

        self.linear = KernelLinear(self.Rs_in, self.Rs_out)

    def __repr__(self):
        return "{name} ({Rs_in} -> {Rs_out})".format(
            name=self.__class__.__name__,
            Rs_in=rs.format_Rs(self.Rs_in),
            Rs_out=rs.format_Rs(self.Rs_out),
        )

    def check_output(self, selection_rule):
        for _, l_out, p_out in self.Rs_out:
            if not any(selection_rule(l_in, p_in, l_out, p_out) for _, l_in, p_in in self.Rs_in):
                raise ValueError("warning! the output (l={}, p={}) cannot be generated".format(l_out, p_out))

    def check_input(self, selection_rule):
        for _, l_in, p_in in self.Rs_in:
            if not any(selection_rule(l_in, p_in, l_out, p_out) for _, l_out, p_out in self.Rs_out):
                raise ValueError("warning! the input (l={}, p={}) cannot be used".format(l_in, p_in))

    def forward(self, r, r_eps=0, **_kwargs):
        """
        :param r: tensor [..., 3]
        :return: tensor [..., l_out * mul_out * m_out, l_in * mul_in * m_in]
        """
        *size, xyz = r.size()
        assert xyz == 3
        r = r.reshape(-1, 3)

        radii = r.norm(2, dim=1)  # [batch]

        # (1) Case r > 0

        # precompute all needed spherical harmonics
        Y = rsh.spherical_harmonics_xyz(self.Ls, r[radii > r_eps])  # [batch, l_filter * m_filter]

        # Normalize the spherical harmonics
        if self.normalization == 'component':
            Y.mul_(math.sqrt(4 * math.pi))
        if self.normalization == 'norm':
            diag = math.sqrt(4 * math.pi) * torch.cat([torch.ones(2 * l + 1) / math.sqrt(2 * l + 1) for _, l, _ in self.Rs_f])
            Y.mul_(diag)

        # use the radial model to fix all the degrees of freedom
        # note: for the normalization we assume that the variance of R[i] is one
        R = self.R(radii[radii > r_eps])  # [batch, l_out * l_in * mul_out * mul_in * l_filter]

        RY = rsh.mul_radial_angular(self.Rs_f, R, Y)

        if Y.shape[0] == 0:
            kernel1 = torch.zeros(0, rs.dim(self.Rs_out), rs.dim(self.Rs_in))
        else:
            kernel1 = self.tp.right(RY)

        # (2) Case r = 0

        kernel2 = self.linear()

        kernel = r.new_zeros(len(r), *kernel2.shape)
        kernel[radii > r_eps] = kernel1
        kernel[radii <= r_eps] = kernel2

        return kernel.reshape(*size, *kernel2.shape)


class FrozenKernel(torch.nn.Module):
    def __init__(self, Rs_in, Rs_out, RadialModel,
                 r, r_eps=0, selection_rule=o3.selection_rule_in_out_sh, normalization='component',
                 allow_unused_inputs=False,
                 allow_zero_outputs=False):
        """
        :param Rs_in: list of triplet (multiplicity, representation order, parity)
        :param Rs_out: list of triplet (multiplicity, representation order, parity)
        :param RadialModel: Class(d), trainable model: R -> R^d
        :param tensor r: [..., 3]
        :param float r_eps: distance considered as zero
        :param selection_rule: function of signature (l_in, p_in, l_out, p_out) -> [l_filter]
        :param sh: spherical harmonics function of signature ([l_filter], xyz[..., 3]) -> Y[m, ...]
        :param normalization: either 'norm' or 'component'
        representation order = nonnegative integer
        parity = 0 (no parity), 1 (even), -1 (odd)
        """
        super().__init__()

        self.Rs_in = rs.convention(Rs_in)
        self.Rs_out = rs.convention(Rs_out)
        if not allow_unused_inputs:
            self.check_input(selection_rule)
        if not allow_zero_outputs:
            self.check_output(selection_rule)

        *self.size, xyz = r.size()
        assert xyz == 3
        r = r.reshape(-1, 3)  # [batch, space]
        self.register_buffer('radii', r.norm(2, dim=1))  # [batch]
        self.r_eps = r_eps

        self.tp = rs.TensorProduct(self.Rs_in, selection_rule, self.Rs_out, normalization, sorted=True)
        self.Rs_f = self.tp.Rs_in2

        Y = rsh.spherical_harmonics_xyz([(1, l, p) for _, l, p in self.Rs_f], r[self.radii > self.r_eps])  # [batch, l_filter * m_filter]

        # Normalize the spherical harmonics
        if normalization == 'component':
            Y.mul_(math.sqrt(4 * math.pi))
        if normalization == 'norm':
            diag = math.sqrt(4 * math.pi) * torch.cat([torch.ones(2 * l + 1) / math.sqrt(2 * l + 1) for _, l, _ in self.Rs_f])
            Y.mul_(diag)

        self.register_buffer('Y', Y)
        self.R = RadialModel(rs.mul_dim(self.Rs_f))

        if (self.radii <= self.r_eps).any():
            self.linear = KernelLinear(self.Rs_in, self.Rs_out)
        else:
            self.linear = None

    def __repr__(self):
        return "{name} ({Rs_in} -> {Rs_out})".format(
            name=self.__class__.__name__,
            Rs_in=rs.format_Rs(self.Rs_in),
            Rs_out=rs.format_Rs(self.Rs_out),
        )

    def check_output(self, selection_rule):
        for _, l_out, p_out in self.Rs_out:
            if not any(selection_rule(l_in, p_in, l_out, p_out) for _, l_in, p_in in self.Rs_in):
                raise ValueError("warning! the output (l={}, p={}) cannot be generated".format(l_out, p_out))

    def check_input(self, selection_rule):
        for _, l_in, p_in in self.Rs_in:
            if not any(selection_rule(l_in, p_in, l_out, p_out) for _, l_out, p_out in self.Rs_out):
                raise ValueError("warning! the input (l={}, p={}) cannot be used".format(l_in, p_in))

    def forward(self):
        """
        :return: tensor [..., l_out * mul_out * m_out, l_in * mul_in * m_in]
        """
        # (1) Case r > 0

        # use the radial model to fix all the degrees of freedom
        # note: for the normalization we assume that the variance of R[i] is one
        R = self.R(self.radii[self.radii > self.r_eps])  # [batch, l_out * l_in * mul_out * mul_in * l_filter]

        RY = rsh.mul_radial_angular(self.Rs_f, R, self.Y)

        if R.shape[0] == 0:
            kernel1 = torch.zeros(0, rs.dim(self.Rs_out), rs.dim(self.Rs_in))
        else:
            kernel1 = self.tp.right(RY)

        # (2) Case r = 0

        if self.linear is not None:
            kernel2 = self.linear()

            kernel = kernel1.new_zeros(len(self.radii), *kernel2.shape)
            kernel[self.radii > self.r_eps] = kernel1
            kernel[self.radii <= self.r_eps] = kernel2
        else:
            kernel = kernel1

        return kernel.reshape(*self.size, *kernel1.shape[1:])



# pylint: disable=missing-docstring, line-too-long, invalid-name, arguments-differ, no-member, pointless-statement
import torch

from e3nn import o3, rs, rsh
from e3nn.kernel import Kernel


class KernelConv(Kernel):
    def forward(self, features, difference_geometry, mask, y=None, radii=None, custom_backward=True):
        """
        :param features: tensor [batch, b, l_in * mul_in * m_in]
        :param difference_geometry: tensor [batch, a, b, xyz]
        :param mask:     tensor [batch, a] (In order to zero contributions from padded atoms.)
        :param y:        Optional precomputed spherical harmonics.
        :param radii:    Optional precomputed normed geometry.
        :param custom_backward: call KernelConvFn rather than using automatic differentiation, (default True)
        :return:         tensor [batch, a, l_out * mul_out * m_out]
        """
        _batch, _a, _b, xyz = difference_geometry.size()
        assert xyz == 3

        if radii is None:
            radii = difference_geometry.norm(2, dim=-1)  # [batch, a, b]

        # precompute all needed spherical harmonics
        if y is None:
            y = rsh.spherical_harmonics_xyz(self.set_of_l_filters, difference_geometry)  # [batch, a, b, l_filter * m_filter]

        y[radii == 0] = 0

        # use the radial model to fix all the degrees of freedom
        # note: for the normalization we assume that the variance of R[i] is one
        r = self.R(radii.flatten()).reshape(*radii.shape, -1)  # [batch, a, b, l_out * l_in * mul_out * mul_in * l_filter]
        r = r.clone()
        r[radii == 0] = 0

        if custom_backward:
            output = KernelConvFn.apply(
                features, y, r, self.norm_coef, self.Rs_in, self.Rs_out, self.selection_rule, self.set_of_l_filters
            )
        else:
            output = kernel_conv_fn_forward(
                features, y, r, self.norm_coef, self.Rs_in, self.Rs_out, self.selection_rule, self.set_of_l_filters
            )

        # Case r > 0
        if radii.shape[1] == radii.shape[2]:
            output += torch.einsum('ij,zaj->zai', self.linear(), features)

        return output * mask.unsqueeze(-1)


def kernel_conv_fn_forward(F, Y, R, norm_coef, Rs_in, Rs_out, selection_rule, set_of_l_filters):
    """
    :param F: tensor [batch, b, l_in * mul_in * m_in]
    :param Y: tensor [l_filter * m_filter, batch, a, b]
    :param R: tensor [batch, a, b, l_out * l_in * mul_out * mul_in * l_filter]
    :param norm_coef: tensor [l_out, l_in]
    :return: tensor [batch, a, l_out * mul_out * m_out, l_in * mul_in * m_in]
    """
    batch, a, b, _ = Y.shape
    n_out = rs.dim(Rs_out)

    kernel_conv = Y.new_zeros(batch, a, n_out)

    # note: for the normalization we assume that the variance of R[i] is one
    begin_R = 0

    begin_out = 0
    for i, (mul_out, l_out, p_out) in enumerate(Rs_out):
        s_out = slice(begin_out, begin_out + mul_out * (2 * l_out + 1))
        begin_out += mul_out * (2 * l_out + 1)

        begin_in = 0
        for j, (mul_in, l_in, p_in) in enumerate(Rs_in):
            s_in = slice(begin_in, begin_in + mul_in * (2 * l_in + 1))
            begin_in += mul_in * (2 * l_in + 1)

            l_filters = selection_rule(l_in, p_in, l_out, p_out)
            if not l_filters:
                continue

            # extract the subset of the `R` that corresponds to the couple (l_out, l_in)
            n = mul_out * mul_in * len(l_filters)
            sub_R = R[:, :, :, begin_R: begin_R + n].reshape(
                batch, a, b, mul_out, mul_in, -1
            )  # [batch, a, b, mul_out, mul_in, l_filter]
            begin_R += n

            K = 0
            for k, l_filter in enumerate(l_filters):
                offset = sum(2 * l + 1 for l in set_of_l_filters if l < l_filter)
                sub_Y = Y[..., offset: offset + 2 * l_filter + 1]  # [batch, a, b, m]

                C = o3.wigner_3j(l_out, l_in, l_filter, cached=True, like=kernel_conv)  # [m_out, m_in, m]

                K += norm_coef[i, j] * torch.einsum(
                    "ijk,zabk,zabuv,zbvj->zaui",
                    C, sub_Y, sub_R[..., k], F[..., s_in].reshape(batch, b, mul_in, -1)
                )  # [batch, a, mul_out, m_out]

            if not isinstance(K, int):
                kernel_conv[:, :, s_out] += K.reshape(batch, a, -1)

    return kernel_conv


class KernelConvFn(torch.autograd.Function):
    """
    The math is presented here:
    https://slides.com/bkmi/convolution-and-kernel/
    """
    @staticmethod
    def forward(ctx, F, Y, R, norm_coef, Rs_in, Rs_out, selection_rule, set_of_l_filters):
        f"""{kernel_conv_fn_forward.__doc__}"""
        ctx.batch, ctx.a, ctx.b, _ = Y.shape
        ctx.Rs_in = Rs_in
        ctx.Rs_out = Rs_out
        ctx.selection_rule = selection_rule
        ctx.set_of_l_filters = set_of_l_filters

        # save necessary tensors for backward
        saved_Y = saved_R = saved_F = None
        if F.requires_grad:
            ctx.F_shape = F.shape
            saved_R = R
            saved_Y = Y
        if Y.requires_grad:
            ctx.Y_shape = Y.shape
            saved_R = R
            saved_F = F
        if R.requires_grad:
            ctx.R_shape = R.shape
            saved_Y = Y
            saved_F = F
        ctx.save_for_backward(saved_F, saved_Y, saved_R, norm_coef)

        return kernel_conv_fn_forward(
            F, Y, R, norm_coef, ctx.Rs_in, ctx.Rs_out, ctx.selection_rule, ctx.set_of_l_filters
        )

    @staticmethod
    def backward(ctx, grad_kernel):  # pragma: no cover
        F, Y, R, norm_coef = ctx.saved_tensors
        batch, a, b = ctx.batch, ctx.a, ctx.b

        grad_F = grad_Y = grad_R = None

        if ctx.needs_input_grad[0]:
            grad_F = grad_kernel.new_zeros(*ctx.F_shape)  # [batch, b, l_in * mul_in * m_in]
        if ctx.needs_input_grad[1]:
            grad_Y = grad_kernel.new_zeros(*ctx.Y_shape)  # [l_filter * m_filter, batch, a, b]
        if ctx.needs_input_grad[2]:
            grad_R = grad_kernel.new_zeros(*ctx.R_shape)  # [batch, a, b, l_out * l_in * mul_out * mul_in * l_filter]

        begin_R = 0

        begin_out = 0
        for i, (mul_out, l_out, p_out) in enumerate(ctx.Rs_out):
            s_out = slice(begin_out, begin_out + mul_out * (2 * l_out + 1))
            begin_out += mul_out * (2 * l_out + 1)

            begin_in = 0
            for j, (mul_in, l_in, p_in) in enumerate(ctx.Rs_in):
                s_in = slice(begin_in, begin_in + mul_in * (2 * l_in + 1))
                begin_in += mul_in * (2 * l_in + 1)

                l_filters = ctx.selection_rule(l_in, p_in, l_out, p_out)
                if not l_filters:
                    continue

                n = mul_out * mul_in * len(l_filters)
                if (grad_Y is not None) or (grad_F is not None):
                    sub_R = R[:, :, :, begin_R: begin_R + n].reshape(
                        batch, a, b, mul_out, mul_in, -1
                    )  # [batch, a, b, mul_out, mul_in, l_filter]
                if grad_R is not None:
                    sub_grad_R = grad_R[:, :, :, begin_R: begin_R + n].clone().reshape(
                        batch, a, b, mul_out, mul_in, -1
                    )  # [batch, a, b, mul_out, mul_in, l_filter]

                if grad_F is not None:
                    sub_grad_F = grad_F[:, :, s_in].clone().reshape(
                        batch, b, mul_in, 2 * l_in + 1
                    )  # [batch, b, mul_in, 2 * l_in + 1]
                if (grad_Y is not None) or (grad_R is not None):
                    sub_F = F[..., s_in].reshape(batch, b, mul_in, 2 * l_in + 1)

                grad_K = grad_kernel[:, :, s_out].reshape(
                    batch, a, mul_out, 2 * l_out + 1
                )

                for k, l_filter in enumerate(l_filters):
                    tmp = sum(2 * l + 1 for l in ctx.set_of_l_filters if l < l_filter)
                    C = o3.wigner_3j(l_out, l_in, l_filter, cached=True, like=grad_kernel)  # [m_out, m_in, m]

                    if (grad_F is not None) or (grad_R is not None):
                        sub_Y = Y[:, :, :, tmp: tmp + 2 * l_filter + 1]  # [batch, a, b, m]

                    if grad_F is not None:
                        sub_grad_F += norm_coef[i, j] * torch.einsum(
                            "zaui,ijk,zabk,zabuv->zbvj",
                            grad_K, C, sub_Y, sub_R[..., k]
                        )  # [batch, b, mul_in, 2 * l_in + 1
                    if grad_Y is not None:
                        grad_Y[..., tmp: tmp + 2 * l_filter + 1] += norm_coef[i, j] * torch.einsum(
                            "zaui,ijk,zabuv,zbvj->zabk",
                            grad_K, C, sub_R[..., k], sub_F
                        )  # [m, batch, a, b]
                    if grad_R is not None:
                        sub_grad_R[..., k] = norm_coef[i, j] * torch.einsum(
                            "zaui,ijk,zabk,zbvj->zabuv",
                            grad_K, C, sub_Y, sub_F
                        )  # [batch, a, b, mul_out, mul_in]
                if grad_F is not None:
                    grad_F[:, :, s_in] = sub_grad_F.reshape(batch, b, mul_in * (2 * l_in + 1))
                if grad_R is not None:
                    grad_R[..., begin_R: begin_R + n] += sub_grad_R.reshape(batch, a, b, -1)
                begin_R += n

        return grad_F, grad_Y, grad_R, None, None, None, None, None



from .cartesian_tensor import CartesianTensor
from .spherical_tensor import SphericalTensor
from .irrep_tensor import IrrepTensor
from .fourier_tensor import FourierTensor


__all__ = ['CartesianTensor', 'SphericalTensor', 'IrrepTensor', 'FourierTensor']



# pylint: disable=C,E1101,E1102
import torch


class torch_default_dtype:

    def __init__(self, dtype):
        self.saved_dtype = None
        self.dtype = dtype

    def __enter__(self):
        self.saved_dtype = torch.get_default_dtype()
        torch.set_default_dtype(self.dtype)

    def __exit__(self, exc_type, exc_value, traceback):
        torch.set_default_dtype(self.saved_dtype)



# pylint: disable=no-member, missing-docstring, invalid-name, redefined-builtin, arguments-differ, line-too-long, unbalanced-tuple-unpacking
import torch

from e3nn import rs


class GatedBlock(torch.nn.Module):
    def __init__(self, Rs_out, scalar_activation, gate_activation):
        """
        :param Rs_out: list of triplet (multiplicity, representation order, parity)
        :param scalar_activation: nonlinear function applied on l=0 channels
        :param gate_activation: nonlinear function applied on the gates
        """
        super().__init__()

        Rs_out = rs.simplify(Rs_out)

        self.scalar_act = scalar_activation
        self.gate_act = gate_activation

        Rs = []
        Rs_gates = []
        for mul, l, p in Rs_out:
            if p != 0:
                raise ValueError("use GatedBlockParity instead")
            Rs.append((mul, l))
            if l != 0:
                Rs_gates.append((mul, 0))

        self.Rs = Rs
        self.Rs_in = rs.simplify(Rs + Rs_gates)

    def forward(self, features, dim=-1):
        """
        :param features: tensor [..., channel, ...]
        :return:         tensor [..., channel, ...]
        """
        dim = (features.dim() + dim) % features.dim()
        size_bef = features.size()[:dim]
        size = features.size(dim)
        size_aft = features.size()[dim + 1:]

        size_out = sum(mul * (2 * l + 1) for mul, l in self.Rs)

        gates = features.narrow(dim, size_out, size - size_out)
        begin_g = 0  # index of first scalar gate capsule

        out = features.new_empty(*size_bef, size_out, *size_aft)
        begin_out = 0  # index of first capsule

        for mul, l in self.Rs:
            if mul == 0:
                continue

            # crop out capsules of order l
            sub = features.narrow(dim, begin_out, mul * (2 * l + 1))  # [..., feature * repr, ...]

            if l == 0:
                # Scalar activation
                sub = self.scalar_act(sub)
            else:
                sub = sub.reshape(*size_bef, mul, 2 * l + 1, *size_aft)  # [..., feature, repr, ...]

                gate = gates.narrow(dim, begin_g, mul)  # [..., feature, ...]
                begin_g += mul

                gate = self.gate_act(gate)
                gate = gate.contiguous().unsqueeze(dim + 1)  # [..., feature, 1, ...]

                sub = (sub * gate).reshape(*size_bef, mul * (2 * l + 1), *size_aft)  # [..., feature * repr, ...]
                del gate

            out.narrow(dim, begin_out, mul * (2 * l + 1)).copy_(sub)
            begin_out += mul * (2 * l + 1)
            del sub

        assert begin_g == size - size_out
        assert begin_out == size_out

        return out



# pylint: disable=not-callable, no-member, invalid-name, line-too-long, wildcard-import, unused-wildcard-import, missing-docstring, bare-except
from setuptools import setup
from pack import KWARGS

setup(**KWARGS)



# pylint: disable=no-member, arguments-differ, redefined-builtin, missing-docstring, line-too-long, invalid-name, abstract-method
from e3nn import rs
from e3nn.non_linearities import GatedBlockParity
from e3nn.non_linearities.rescaled_act import sigmoid, swish, tanh


def make_gated_block(Rs_in, mul=16, lmax=3):
    scalars = [(mul, l, p) for mul, l, p in [(mul, 0, +1), (mul, 0, -1)] if rs.haslinearpath(Rs_in, l, p)]
    act_scalars = [(mul, swish if p == 1 else tanh) for mul, l, p in scalars]

    nonscalars = [(mul, l, p) for l in range(1, lmax + 1) for p in [+1, -1] if rs.haslinearpath(Rs_in, l, p)]
    if rs.haslinearpath(Rs_in, 0, +1):
        gates = [(rs.mul_dim(nonscalars), 0, +1)]
        act_gates = [(-1, sigmoid)]
    else:
        gates = [(rs.mul_dim(nonscalars), 0, -1)]
        act_gates = [(-1, tanh)]

    return GatedBlockParity(scalars, act_scalars, gates, act_gates, nonscalars)



# pylint: disable=no-member, arguments-differ, missing-docstring, invalid-name, line-too-long
import torch
import torch.nn as nn


class BatchNorm(nn.Module):
    def __init__(self, Rs, eps=1e-5, momentum=0.1, affine=True, reduce='mean', normalization='component'):
        '''
        Batch normalization layer for orthonormal representations
        It normalizes by the norm of the representations.
        Not that the norm is invariant only for orthonormal representations.
        Irreducible representations `o3.irr_repr` are orthonormal.

        input shape : [batch, [spacial dimensions], stacked orthonormal representations]

        :param Rs: list of tuple (multiplicity, dimension)
        :param eps: avoid division by zero when we normalize by the variance
        :param momentum: momentum of the running average
        :param affine: do we have weight and bias parameters
        :param reduce: method to contract over the spacial dimensions
        '''
        super().__init__()

        self.Rs = [(m, d) for m, d in Rs if m * d > 0]
        self.eps = eps
        self.momentum = momentum
        self.affine = affine

        num_scalar = sum(m for m, d in Rs if d == 1)
        num_features = sum(m for m, d in Rs)

        self.register_buffer('running_mean', torch.zeros(num_scalar))
        self.register_buffer('running_var', torch.ones(num_features))

        if affine:
            self.weight = nn.Parameter(torch.ones(num_features))
            self.bias = nn.Parameter(torch.zeros(num_scalar))
        else:
            self.register_parameter('weight', None)
            self.register_parameter('bias', None)

        assert isinstance(reduce, str), "reduce should be passed as a string value"
        assert reduce in ['mean', 'max'], "reduce needs to be 'mean' or 'max'"
        self.reduce = reduce

        assert normalization in ['norm', 'component'], "normalization needs to be 'norm' or 'component'"
        self.normalization = normalization

    def __repr__(self):
        return "{} (Rs={}, eps={}, momentum={})".format(
            self.__class__.__name__,
            self.Rs,
            self.eps,
            self.momentum)

    def _roll_avg(self, curr, update):
        return (1 - self.momentum) * curr + self.momentum * update.detach()

    def forward(self, input):  # pylint: disable=redefined-builtin
        '''
        :param input: [batch, ..., stacked features]
        '''
        batch, *size, dim = input.shape
        input = input.reshape(batch, -1, dim)  # [batch, sample, stacked features]

        if self.training:
            new_means = []
            new_vars = []

        fields = []
        ix = 0
        irm = 0
        irv = 0
        iw = 0
        ib = 0

        for m, d in self.Rs:
            field = input[:, :, ix: ix + m * d]  # [batch, sample, mul * repr]
            ix += m * d

            # [batch, sample, mul, repr]
            field = field.reshape(batch, -1, m, d)

            if d == 1:  # scalars
                if self.training:
                    field_mean = field.mean([0, 1]).reshape(m)  # [mul]
                    new_means.append(
                        self._roll_avg(self.running_mean[irm:irm + m], field_mean)
                    )
                else:
                    field_mean = self.running_mean[irm: irm + m]
                irm += m

                # [batch, sample, mul, repr]
                field = field - field_mean.reshape(m, 1)

            if self.training:
                if self.normalization == 'norm':
                    field_norm = field.pow(2).sum(3)  # [batch, sample, mul]
                elif self.normalization == 'component':
                    field_norm = field.pow(2).mean(3)  # [batch, sample, mul]
                else:
                    raise ValueError("Invalid normalization option {}".format(self.normalization))

                if self.reduce == 'mean':
                    field_norm = field_norm.mean(1)  # [batch, mul]
                elif self.reduce == 'max':
                    field_norm = field_norm.max(1).values  # [batch, mul]
                else:
                    raise ValueError("Invalid reduce option {}".format(self.reduce))

                field_norm = field_norm.mean(0)  # [mul]
                new_vars.append(self._roll_avg(self.running_var[irv: irv + m], field_norm))
            else:
                field_norm = self.running_var[irv: irv + m]
            irv += m

            field_norm = (field_norm + self.eps).pow(-0.5)  # [mul]

            if self.affine:
                weight = self.weight[iw: iw + m]  # [mul]
                iw += m

                field_norm = field_norm * weight  # [mul]

            field = field * field_norm.reshape(m, 1)  # [batch, sample, mul, repr]

            if self.affine and d == 1:  # scalars
                bias = self.bias[ib: ib + m]  # [mul]
                ib += m
                field += bias.reshape(m, 1)  # [batch, sample, mul, repr]

            fields.append(field.reshape(batch, -1, m * d))  # [batch, sample, mul * repr]

        if ix != dim:
            fmt = "`ix` should have reached input.size(-1) ({}), but it ended at {}"
            msg = fmt.format(dim, ix)
            raise AssertionError(msg)

        if self.training:
            assert irm == self.running_mean.numel()
            assert irv == self.running_var.size(0)
        if self.affine:
            assert iw == self.weight.size(0)
            assert ib == self.bias.numel()

        if self.training:
            self.running_mean.copy_(torch.cat(new_means))
            self.running_var.copy_(torch.cat(new_vars))

        output = torch.cat(fields, dim=2)  # [batch, sample, stacked features]
        return output.reshape(batch, *size, dim)



# pylint: disable=not-callable, no-member, invalid-name, line-too-long, arguments-differ
"""
Real Spherical Harmonics equivariant with respect to o3.rot and o3.irr_repr
"""
import math
from functools import lru_cache

import torch
from sympy import Integer, Poly, diff, factorial, pi, sqrt, symbols

from e3nn import rs
from e3nn.util.eval_code import eval_code


def spherical_harmonics_expand_matrix(ls, like=None):
    """
    convertion matrix between a flatten vector (L, m) like that
    (0, 0) (1, -1) (1, 0) (1, 1) (2, -2) (2, -1) (2, 0) (2, 1) (2, 2)

    and a bidimensional matrix representation like that
                    (0, 0)
            (1, -1) (1, 0) (1, 1)
    (2, -2) (2, -1) (2, 0) (2, 1) (2, 2)

    :return: tensor [l, m, l * m]
    """
    lmax = max(ls)
    zeros = torch.zeros if like is None else like.new_zeros
    m = zeros(len(ls), 2 * lmax + 1, sum(2 * l + 1 for l in ls))
    i = 0
    for j, l in enumerate(ls):
        m[j, lmax - l: lmax + l + 1, i:i + 2 * l + 1] = torch.eye(2 * l + 1)
        i += 2 * l + 1
    return m


@torch.jit.script
def mul_m_lm(Rs: rs.TY_RS_STRICT, x_m: torch.Tensor, x_lm: torch.Tensor) -> torch.Tensor:  # pragma: no cover
    """
    multiply tensor [..., l * m] by [..., m]
    """
    lmax = x_m.shape[-1] // 2
    out = []
    i = 0
    for mul, l, _ in Rs:
        d = mul * (2 * l + 1)
        x1 = x_lm[..., i: i + d]  # [..., mul * m]
        x1 = x1.reshape(x1.shape[:-1] + (mul, 2 * l + 1))  # [..., mul, m]
        x2 = x_m[..., lmax - l: lmax + l + 1]  # [..., m]
        x2 = x2.reshape(x2.shape[:-1] + (1, 2 * l + 1))  # [..., mul=1, m]
        x = x1 * x2
        x = x.reshape(x.shape[:-2] + (d,))
        out.append(x)
        i += d
    return torch.cat(out, dim=-1)


@torch.jit.script
def mul_radial_angular(Rs: rs.TY_RS_STRICT, radial, angular):  # pragma: no cover
    """
    :param Rs: output representation
    :param angular: [..., l * m]
    :param radial: [..., l * mul]
    """
    n = 0
    for mul, l, _ in Rs:
        n += mul * (2 * l + 1)

    y = radial[..., 0] * angular[..., 0]
    out = radial.new_empty(y.shape + (n,))

    a = 0
    r = 0
    i = 0
    for mul, l, _ in Rs:
        dim = mul * (2 * l + 1)
        x = radial[..., r: r + mul, None] * angular[..., None, a: a + 2 * l + 1]
        x = x.reshape(y.shape + (dim,))
        out[..., i: i + dim] = x
        i += dim
        r += mul
        a += 2 * l + 1

    assert r == radial.shape[-1]
    assert a == angular.shape[-1]

    return out


def sympy_legendre(l, m):
    """
    en.wikipedia.org/wiki/Associated_Legendre_polynomials
    - remove two times (-1)^m
    - use another normalization such that P(l, -m) = P(l, m)

    y = sqrt(1 - z^2)
    """
    l = Integer(l)
    m = Integer(abs(m))
    z, y = symbols('z y', real=True)
    ex = 1 / (2**l * factorial(l)) * y**m * diff((z**2 - 1)**l, z, l + m)
    ex *= (-1)**l * sqrt((2 * l + 1) / (4 * pi) * factorial(l - m) / factorial(l + m))
    return ex


def poly_legendre(l, m):
    """
    polynomial coefficients of legendre

    y = sqrt(1 - z^2)
    """
    z, y = symbols('z y', real=True)
    return Poly(sympy_legendre(l, m), domain='R', gens=(z, y)).as_dict()


_legendre_code = """
import torch
from e3nn import rsh

@torch.jit.script
def main(z: torch.Tensor, y: torch.Tensor) -> torch.Tensor:
    out = z.new_zeros(z.shape + (lsize,))

# fill out

    return out
"""


@lru_cache()
def _legendre_genjit(ls):
    ls = list(ls)
    fill = ""
    i = 0
    for l in ls:
        for m in range(l + 1):
            p = poly_legendre(l, m)
            formula = " + ".join("{:.25f} * z**{} * y**{}".format(c, zn, yn) for (zn, yn), c in p.items())
            fill += "    l{} = {}\n".format(m, formula)

        for m in range(-l, l + 1):
            fill += "    out[..., {}] = l{}\n".format(i, abs(m))
            i += 1

    code = _legendre_code
    code = code.replace("lsize", str(sum(2 * l + 1 for l in ls)))
    code = code.replace("# fill out", fill)
    return eval_code(code).main


def legendre(ls, z, y=None):
    """
    associated Legendre polynomials

    :param ls: list
    :param z: tensor of shape [...]
    :return: tensor of shape [..., l * m]
    """
    if y is None:
        y = (1 - z**2).relu().sqrt()

    return _legendre_genjit(tuple(ls))(z, y)


def spherical_harmonics_z(Rs, z, y=None):
    """
    the z component of the spherical harmonics
    (useful to perform fourier transform)

    :param z: tensor of shape [...]
    :return: tensor of shape [..., l * m]
    """
    Rs = rs.simplify(Rs)
    assert all(p in [0, (-1)**l] for _, l, p in Rs)
    ls = [l for mul, l, _ in Rs for _ in range(mul)]
    return legendre(ls, z, y)  # [..., l * m]


@torch.jit.script
def spherical_harmonics_alpha(l: int, alpha: torch.Tensor) -> torch.Tensor:  # pragma: no cover
    """
    the alpha (x, y) component of the spherical harmonics
    (useful to perform fourier transform)

    :param alpha: tensor of shape [...]
    :return: tensor of shape [..., m]
    """
    alpha = alpha.unsqueeze(-1)  # [..., 1]

    m = torch.arange(1, l + 1, dtype=alpha.dtype, device=alpha.device)  # [1, 2, 3, ..., l]
    cos = torch.cos(m * alpha)  # [..., m]

    m = torch.arange(l, 0, -1, dtype=alpha.dtype, device=alpha.device)  # [l, l-1, l-2, ..., 1]
    sin = torch.sin(m * alpha)  # [..., m]

    out = torch.cat([
        math.sqrt(2) * sin,
        torch.ones_like(alpha),
        math.sqrt(2) * cos,
    ], dim=alpha.ndim-1)

    return out  # [..., m]


def spherical_harmonics_alpha_beta(Rs, alpha, beta):
    """
    spherical harmonics

    :param Rs: list of L's
    :param alpha: float or tensor of shape [...]
    :param beta: float or tensor of shape [...]
    :return: tensor of shape [..., m]
    """
    if alpha.device.type == 'cuda' and beta.device.type == 'cuda' and not alpha.requires_grad and not beta.requires_grad and rs.lmax(Rs) <= 10:  # pragma: no cover
        xyz = torch.stack([beta.sin() * alpha.cos(), beta.sin() * alpha.sin(), beta.cos()], dim=-1)
        try:
            return _spherical_harmonics_xyz_cuda(Rs, xyz)
        except ImportError:
            pass

    return spherical_harmonics_alpha_z_y(Rs, alpha, beta.cos(), beta.sin().abs())


def spherical_harmonics_alpha_z_y(Rs, alpha, z, y):
    """
    cpu version of spherical_harmonics_alpha_beta
    """
    Rs = rs.simplify(Rs)
    sha = spherical_harmonics_alpha(rs.lmax(Rs), alpha.flatten())  # [z, m]
    shz = spherical_harmonics_z(Rs, z.flatten(), y.flatten())  # [z, l * m]
    out = mul_m_lm(Rs, sha, shz)
    return out.reshape(alpha.shape + (shz.shape[1],))


@lru_cache()
def _rep_zx(Rs, dtype, device):
    o = torch.zeros((), dtype=dtype, device=device)
    return rs.rep(Rs, o, -math.pi / 2, o)


def spherical_harmonics_xyz(Rs, xyz, normalization='none'):
    """
    spherical harmonics

    :param Rs: list of L's
    :param xyz: tensor of shape [..., 3]
    :return: tensor of shape [..., m]
    """
    Rs = rs.simplify(Rs)
    *size, _ = xyz.shape
    xyz = xyz.reshape(-1, 3)
    d = torch.norm(xyz, 2, dim=1)
    xyz = xyz[d > 0]
    xyz = xyz / d[d > 0, None]

    sh = None

    if xyz.device.type == 'cuda' and not xyz.requires_grad and rs.lmax(Rs) <= 10:  # pragma: no cover
        try:
            sh = _spherical_harmonics_xyz_cuda(Rs, xyz)
        except ImportError:
            pass

    if sh is None:
        # if z > x, rotate x-axis with z-axis
        s = xyz[:, 2].abs() > xyz[:, 0].abs()
        xyz[s] = xyz[s] @ xyz.new_tensor([[0, 0, 1], [0, 1, 0], [-1, 0, 0]])

        alpha = torch.atan2(xyz[:, 1], xyz[:, 0])
        z = xyz[:, 2]
        y = xyz[:, :2].norm(dim=1)

        sh = spherical_harmonics_alpha_z_y(Rs, alpha, z, y)

        # rotate back
        sh[s] = sh[s] @ _rep_zx(tuple(Rs), xyz.dtype, xyz.device)

    if len(d) > len(sh):
        out = sh.new_zeros(len(d), sh.shape[1])
        out[d == 0] = math.sqrt(1 / (4 * math.pi)) * torch.cat([sh.new_ones(1) if l == 0 else sh.new_zeros(2 * l + 1) for mul, l, p in Rs for _ in range(mul)])
        out[d > 0] = sh
        sh = out

    if normalization == 'component':
        sh.mul_(math.sqrt(4 * math.pi))
    if normalization == 'norm':
        sh.mul_(torch.cat([math.sqrt(4 * math.pi / (2 * l + 1)) * sh.new_ones(2 * l + 1) for mul, l, p in Rs for _ in range(mul)]))
    return sh.reshape(*size, sh.shape[1])


def _spherical_harmonics_xyz_cuda(Rs, xyz):  # pragma: no cover
    """
    cuda version of spherical_harmonics_xyz
    """
    from e3nn import cuda_rsh  # pylint: disable=no-name-in-module, import-outside-toplevel

    lmax = rs.lmax(Rs)
    out = xyz.new_empty(((lmax + 1)**2, xyz.size(0)))  # [ filters, batch_size]
    cuda_rsh.real_spherical_harmonics(out, xyz)

    # (-1)^L same as (pi-theta) -> (-1)^(L+m) and 'quantum' norm (-1)^m combined  # h - halved
    norm_coef = [elem for lh in range((lmax + 1) // 2) for elem in [1.] * (4 * lh + 1) + [-1.] * (4 * lh + 3)]
    if lmax % 2 == 0:
        norm_coef.extend([1.] * (2 * lmax + 1))
    norm_coef = out.new_tensor(norm_coef).unsqueeze(1)
    out.mul_(norm_coef)

    if not rs.are_equal(Rs, list(range(lmax + 1))):
        out = torch.cat([out[l**2: (l + 1)**2] for mul, l, _ in Rs for _ in range(mul)])

    return out.T



'''
Cache in files
'''
import fcntl
import glob
import gzip
import os
import pickle
import sys
from functools import lru_cache, wraps
from itertools import chain, count


class FileSystemMutex:
    '''
    Mutual exclusion of different **processes** using the file system
    '''

    def __init__(self, filename):
        self.handle = None
        self.filename = filename

    def acquire(self):
        '''
        Locks the mutex
        if it is already locked, it waits (blocking function)
        '''
        self.handle = open(self.filename, 'w')
        fcntl.lockf(self.handle, fcntl.LOCK_EX)
        self.handle.write("{}\n".format(os.getpid()))
        self.handle.flush()

    def release(self):
        '''
        Unlock the mutex
        '''
        if self.handle is None:
            raise RuntimeError()
        fcntl.lockf(self.handle, fcntl.LOCK_UN)
        self.handle.close()
        self.handle = None

    def __enter__(self):
        self.acquire()

    def __exit__(self, exc_type, exc_value, traceback):
        self.release()


def cached_picklesjar(dirname, maxsize=128, open_jar=gzip.open,
                      load=pickle.load, save=pickle.dump, ext='pickle'):
    '''
    Cache a function with a directory

    :param dirname: the directory path
    :param maxsize: maximum size of the RAM cache (there is no limit for the directory cache)
    '''

    def decorator(func):
        '''
        The actual decorator
        '''

        @lru_cache(maxsize=maxsize)
        @wraps(func)
        def wrapper(*args, **kwargs):
            '''
            The wrapper of the function
            '''
            try:
                os.makedirs(dirname)
            except FileExistsError:
                pass
            except PermissionError:
                return func(*args, **kwargs)

            if not os.access(dirname, os.W_OK):
                return func(*args, **kwargs)

            mutexfile = os.path.join(dirname, "mutex")

            key = (args, frozenset(kwargs.items()), func.__defaults__)

            with FileSystemMutex(mutexfile):
                for file in glob.glob(os.path.join(dirname, "*.{}".format(ext))):
                    with open_jar(file, "rb") as file:
                        loadedkey = load(file)
                        if key == loadedkey:
                            return load(file)

            sys.stdout.flush()
            result = func(*args, **kwargs)
            sys.stdout.flush()

            with FileSystemMutex(mutexfile):
                name = " ".join(map(str, args))
                if kwargs:
                    name += " " + " ".join(
                        "{}={}".format(key, value)
                        for key, value in sorted(kwargs.items())
                    )

                for postfix in chain([''], ('_{}'.format(i) for i in count())):
                    file = os.path.join(dirname, "{}{}.{}".format(name, postfix, ext))
                    if not os.path.isfile(file):
                        break

                try:
                    with open_jar(file, "wb") as file:
                        save(key, file)
                        save(result, file)
                except PermissionError:
                    pass

            return result

        return wrapper

    return decorator
